from cowboy_lib.utils import locate_python_interpreter

from pathlib import Path
import subprocess
from datetime import datetime, timedelta
import subprocess


class Manager:
    """
    Interacts with client running in background
    """

    def __init__(
        self, heart_beat_fp: Path, heart_beat_interval: int = 5, console: bool = False
    ):
        self.heart_beat_fp = heart_beat_fp
        self.heart_beat_interval = heart_beat_interval
        self.console = console
        self.interp = locate_python_interpreter()

        if not self.is_alive():
            # print("Client not alive starting client")
            self.start_client()
        # else:
        # print("Client is alive!")

    def start_client(self):
        subprocess.Popen(
            [
                self.interp,
                "-m",
                "cowboy.task_client.client",
                str(self.heart_beat_fp),
                str(self.heart_beat_interval),
                "True",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

    def is_alive(self):
        """
        Checks that client is still alive if last beat is within
        time interval
        """
        if not self.read_beat():
            return False

        # adding one to the interval to account lag
        if datetime.now() - self.read_beat() < timedelta(
            seconds=self.heart_beat_interval + 1
        ):
            return True

        return False

    def read_beat(self):
        try:
            with open(self.heart_beat_fp, "r") as f:
                hb_time = f.readlines()[-1].strip()

                return datetime.strptime(hb_time, "%Y-%m-%d %H:%M:%S")

        except ValueError as e:
            print(
                f"Possibly corrupted file: {str(self.heart_beat_fp.resolve())}, try deleting "
            )
        except FileNotFoundError:
            return None



from cowboy.http import APIClient
from cowboy.db.core import Database

from typing import List

db = Database()
api = APIClient(db)


### stopped here
def api_experiment(
    repo_name: str,
    test_modules: List[str],
):
    """
    Augments existing test modules with new test cases
    """
    pass
    #
    # response = api.long_post(
    #     "/test-gen/augment",
    #     {
    #         "src_file": src_file,
    #         "repo_name": repo_name,
    #         "mode": mode,
    #         "tms": tms,
    #     },
    # )

    # return response



from cowboy_lib.utils import generate_id
from cowboy_lib.ast.code import Function
from cowboy_lib.repo.repository import PatchFile

from pydantic import BaseModel, Field, ConfigDict, model_validator
from typing import List, Optional, Any, Tuple, Dict
from enum import Enum
from pathlib import Path


class TaskStatus(Enum):
    PENDING = "PENDING"
    STARTED = "STARTED"
    COMPLETE = "COMPLETE"
    FAILED = "FAILED"


class TaskType(str, Enum):
    SHUTDOWN = "SHUTDOWN"
    RUN_TEST = "RUN_TEST"


class TaskResult(BaseModel):
    coverage: Optional[Dict] = None
    failed: Optional[Dict] = None
    exception: Optional[str] = None

    @model_validator(mode="before")
    def check_coverage_or_exception(cls, values):
        coverage, failed, exception = (
            values.get("coverage"),
            values.get("failed"),
            values.get("exception"),
        )
        if exception and (coverage or failed):
            raise ValueError(
                "If 'exception' is specified, 'coverage' and 'failed' must not be specified."
            )
        if not exception and not (coverage or failed):
            raise ValueError(
                "Either 'coverage' and 'failed' or 'exception' must be specified."
            )
        return values


class Task(BaseModel):
    """
    Task datatype
    """

    type: TaskType
    task_id: str = Field(default_factory=lambda: generate_id())
    result: Optional[TaskResult] = Field(default=None)
    status: str = Field(default=TaskStatus.PENDING.value)
    task_args: Optional[Any]


class FunctionArg(BaseModel):
    name: str
    is_meth: bool


# each classmethod below accepts different signatures which are inconsistent
# with the class fields of the Pydantic BaseModel, so we had to set the
# class fields to Any... need better way of expressing this
class RunTestTaskArgs(BaseModel):
    repo_name: str
    patch_file: Optional[Any] = None
    exclude_tests: List[Tuple[Any, Any]] = Field(default_factory=list)
    include_tests: List[str] = Field(default_factory=list)

    @classmethod
    def from_data(
        cls,
        repo_name: str,
        exclude_tests: List[Tuple[Tuple[str, bool], str]] = [],
        include_tests: List[str] = [],
        patch_file: PatchFile = None,
    ):
        """
        Used by server
        """
        partial = cls(
            repo_name=repo_name,
            patch_file=patch_file,
            exclude_tests=exclude_tests,
            include_tests=include_tests,
        )

        if partial.exclude_tests:
            partial.exclude_tests = [
                (
                    FunctionArg(
                        name=func[0],
                        is_meth=func[1],
                    ),
                    str(path),
                )
                for func, path in partial.exclude_tests
            ]

        return partial

    @classmethod
    def from_json(
        cls,
        repo_name: str,
        patch_file: Dict = {},
        exclude_tests: List[Tuple[Dict, str]] = [],
        include_tests: List[str] = [],
    ):
        """
        Used by client
        """
        partial = cls(
            repo_name=repo_name,
            patch_file=patch_file,
            exclude_tests=exclude_tests,
            include_tests=include_tests,
        )

        if partial.patch_file:
            partial.patch_file = PatchFile(
                path=Path(partial.patch_file["path"]),
                patch=partial.patch_file["patch"],
            )
        if partial.exclude_tests:
            partial.exclude_tests = [
                (
                    FunctionArg(
                        name=func[0],
                        is_meth=func[1],
                    ),
                    Path(path),
                )
                for func, path in partial.exclude_tests
            ]

        return partial



from pathlib import Path
import subprocess
import os
import uuid
import random
import string
import shutil


def get_current_git_commit(repo_path: Path) -> str:
    """
    Uses subprocess to get the current git commit hash.

    Returns:
        str: The current git commit hash.
    """
    try:
        commit_hash = (
            subprocess.check_output(
                ["cd", str(repo_path.resolve()), "&&", "git", "rev-parse", "HEAD"],
                shell=True,
            )
            .strip()
            .decode("utf-8")
        )
        return commit_hash
    except subprocess.CalledProcessError as e:
        print(f"Error getting current git commit: {e}")
        return ""


def locate_python_interpreter():
    possible_interpreters = ["python", "python3"]

    for interpreter in possible_interpreters:
        # Check if the interpreter is in PATH
        path = shutil.which(interpreter)
        if path:
            return path

    # Try manually common locations
    common_locations = [
        "/usr/bin/python",
        "/usr/local/bin/python",
        "/usr/bin/python3",
        "/usr/local/bin/python3",
        "/bin/python",
        "/bin/python3",
        "/usr/sbin/python",
        "/usr/sbin/python3",
    ]

    for location in common_locations:
        if os.path.isfile(location) and os.access(location, os.X_OK):
            return location

    # As a last resort, try running "python --version" and "python3 --version"
    for interpreter in possible_interpreters:
        try:
            result = subprocess.run(
                [interpreter, "--version"], capture_output=True, text=True
            )
            if result.returncode == 0:
                return interpreter
        except (subprocess.CalledProcessError, FileNotFoundError):
            continue

    raise FileNotFoundError("Python interpreter not found on this host")


def testfiles_in_coverage(base_cov, src_repo) -> bool:
    """
    Check if the test files are accidentally included in the coverage
    """
    for test_file in src_repo.test_files:
        for cov in base_cov.cov_list:
            if cov.filename.split(os.sep)[-1] == test_file.path.name:
                return True
    return False


def generate_id():
    """
    Generates a random UUID
    """
    return str(uuid.uuid4())


def gen_random_name():
    """
    Generates a random name using ASCII, 8 characters in length
    """

    return "".join(random.choices(string.ascii_lowercase, k=8))


def find_substring(string: str, substring: str):
    indices = []
    start = 0
    while start < len(string):
        start = string.find(substring, start)
        if start == -1:  # No more occurrences
            return indices
        indices.append(start)
        start += 1  # Move past the last found index to find subsequent matches

    return indices



from typing import List, Tuple
import ast

from ..code import Function, Class, Decorator, Argument, ASTNode, NodeType

from logging import getLogger

logger = getLogger("test_results")
AST_FUNCTIONS = (ast.FunctionDef, ast.AsyncFunctionDef)


# REFACTOR-AST: this whole function needs to be replaced by the Bloop/tree-sitter
# implementation
class PythonAST:
    def __init__(self, code: str):
        assert isinstance(code, str)

        self.classes: List[ASTNode] = []
        self.functions: List[ASTNode] = []

        self._code = code

    # Thank you GPT!
    def _parse_decorators(self, node):
        decorators = []
        for d in node.decorator_list:
            # Initialize d_name as None
            d_name = None

            # Handle direct attribute access or function call with attribute access
            if isinstance(d, ast.Attribute) or (
                isinstance(d, ast.Call) and isinstance(d.func, ast.Attribute)
            ):
                d_name_parts = []

                # If it's a Call, we start with d.func to get the Attribute node
                current_node = d.func if isinstance(d, ast.Call) else d

                # Traverse the Attribute nodes
                while isinstance(current_node, ast.Attribute):
                    d_name_parts.append(current_node.attr)
                    current_node = current_node.value

                # The loop ends at an ast.Name node, which gives us the root name
                if isinstance(current_node, ast.Name):
                    d_name_parts.append(current_node.id)

                # Combine the parts to form the full decorator name
                d_name = ".".join(reversed(d_name_parts))

            elif isinstance(d, ast.Name):
                d_name = d.id

            # Append the decorator name to the list if it was found
            if d_name:
                range = self._get_range(d)
                decorators.append(
                    Decorator(
                        d_name,
                        # bring range to 0-based indexing
                        (range[0], getattr(d, "end_lineno", range[0]) - 1),
                        None,
                        [],
                        self._get_lines(*range),
                        d,
                    )
                )

        return decorators

    def _parse_classes(self, child: ast.AST, parent: ast.AST):
        classes = []
        if isinstance(child, ast.ClassDef):
            range = self._get_range(child)
            decorators = self._parse_decorators(child)

            scope = None
            if isinstance(parent, AST_FUNCTIONS) and isinstance(parent, ast.ClassDef):
                scope = self.find_node(parent)

            classes.append(
                Class(
                    child.name, range, scope, decorators, self._get_lines(*range), child
                )
            )
        return classes

    def _parse_functions(self, child: ast.AST, parent: ast.AST):
        functions = []
        if isinstance(child, AST_FUNCTIONS):
            func_name = child.name
            child_range = self._get_range(child)
            decorators = self._parse_decorators(child)

            args = [
                Argument(arg.arg, getattr(arg.annotation, "id", None))
                for arg in child.args.args
            ]

            # rethink this loop when not hungover...
            # assign funcs and classes as parents
            scope = None
            if isinstance(parent, AST_FUNCTIONS) or isinstance(parent, ast.ClassDef):
                scope = self.find_node(parent)

            func = Function(
                func_name,
                child_range,
                scope,
                decorators,
                self._get_lines(*child_range),
                child,
                arguments=args,
            )

            # if a parent class contains a test function, then parent class is a test class
            if scope and isinstance(parent, ast.ClassDef):
                scope.add_func(func)
                if func.is_test:
                    scope.set_is_test(True)

            functions.append(func)

        return functions

    def find_node(self, node: ast.AST) -> ASTNode:
        return next(filter(lambda x: x.is_ast(node), self.classes + self.functions))

    def _get_range(self, node: ast.AST):
        """AST nodes are 1-indexed"""
        return (node.lineno - 1, node.end_lineno - 1)

    def _get_lines(self, start: int, end: int) -> List[str]:
        return self._code.split("\n")[start : end + 1]

    def parse(self) -> Tuple[List[Function], List[Class]]:
        def set_parents(node, parent=None):
            for child in ast.iter_child_nodes(node):
                child.parent = node
                set_parents(child, node)

        def parse_ast(node, parent=None):
            # classes guranteed to be parsed before their child methods right?
            for child in ast.iter_child_nodes(node):
                self.classes.extend(self._parse_classes(child, parent))
                self.functions.extend(self._parse_functions(child, parent))

                parse_ast(child, parent=child)

        tree = ast.parse(self._code)

        set_parents(tree)
        parse_ast(tree)

        return self.functions, self.classes



from cowboy.config import REACT_CONFIG_PATH
from pathlib import Path
import json


def init_react_env_vars(token, api_endpoint):
    """
    Init the .env file in the react folder
    """
    env_vars = {
        "jwt_token": token,
        "api_endpoint": api_endpoint,
    }

    with open(REACT_CONFIG_PATH, "w") as f:
        f.write(json.dumps(env_vars))



from abc import ABC, abstractmethod
from cowboy_lib.api.runner.shared import RunTestTaskArgs
from cowboy_lib.coverage import CoverageResult

from typing import Tuple, List, Any


class RunnerError(Exception):
    def __init__(self, *args):
        super().__init__(*args)

        self.args = args

    def __str__(self):
        msg = "\n".join(self.args)
        return f"RunnerException => {msg}"


class TestSuiteError(RunnerError):
    """
    Stderr from the subprocess that runs the lang/framework unit test suite
    """

    def __init__(self, stderr: str, cloned_path: str, cmd_str: str):
        msg = [
            f"TestSuiteError: ({cloned_path})",
            f"CMD: {cmd_str}",
            stderr,
        ]
        super().__init__(*msg)

        # self.stderr = stderr
        # self.cloned_path = cloned_path
        # self.cmd_str = cmd_str

    # Not actually used for some reason
    # def __str__(self):
    #     msg = [
    #         f"TestSuiteError: ({self.cloned_path})",
    #         f"CMD: {self.cmd_str}",
    #         self.stderr,
    #     ]

    #     return super().__str__(*msg)


class Runner(ABC):
    """
    Runs the lang/framework specific unit test
    """

    @abstractmethod
    def run_testsuite(self, args: RunTestTaskArgs) -> Tuple[CoverageResult, str, str]:
        """
        Runs the lang/framework specific unit test suite
        """
        raise NotImplementedError

    def _construct_cmd(
        self,
        repo_path,
        selected_args_str: str = "",
        deselected_args_str: str = "",
    ):
        """
        Constructs the cmd for running the test via subprocess
        """
        raise NotImplementedError

    def _get_include_tests_arg_str(self, include_tests: List[str] = []):
        """
        Constructs the arg string for selecting specific tests
        """
        raise NotImplementedError

    def _get_exclude_tests_arg_str(self, exclude_tests: List[Any]):
        """
        Constructs the arg string for excluding specific tests
        """
        raise NotImplementedError



from cowboy_lib.repo.source_file import TestFile, SourceFile, Function, Class

from pathlib import Path
from functools import reduce
from typing import List, Optional, Iterator, Tuple, TYPE_CHECKING

from logging import getLogger

logger = getLogger("test_results")
longterm_logger = getLogger("longterm")


# TODO: should we combine git/src_repo into one?
class SourceRepo:
    """
    Used by the TestStrategy to access files and their contents
    """

    def __init__(self, repo_path: Path, files_list: List[str] = []):
        self.repo_path = repo_path
        self.files_list = files_list
        self.source_files: List[SourceFile] = self._init_source_files(files_list)
        # counter = Counter([f.path for f in self.source_files]).most_common()
        # print(counter)

    def get_rel_path(self, file_path: Path) -> str:
        """
        Get the path relative to the source repo
        """
        return file_path.relative_to(self.repo_path)

    @property
    def test_files(self) -> List[TestFile]:
        return [f for f in self.source_files if isinstance(f, TestFile)]

    def _init_source_files(self, files_list=[]) -> List[TestFile]:
        """
        Finds all test files in the repo
        """
        source_files = []
        for path in self.repo_path.rglob("*"):
            if path.is_file() and path.name.endswith(".py"):
                if files_list:
                    if path not in [self.repo_path / f for f in files_list]:
                        continue
                with open(path, "r", encoding="utf-8") as f:
                    lines = f.read().split("\n")
                try:
                    rel_path = self.get_rel_path(path)
                    source_file = (
                        TestFile(lines, rel_path)
                        if path.name.startswith("test_")
                        else SourceFile(lines, rel_path)
                    )
                    source_files.append(source_file)
                except SyntaxError as e:
                    logger.error(f"AST Syntax error while parsing: {path}")

        return source_files

    def find_node(self, name: str, file: str, node_type: str) -> Optional[Function]:
        """
        Finds a function in a file. A node is uniquely identified by its name and filepath
        (local scope)
        """
        for source_file in self.source_files:
            # if str(source_file.path) == file: doesnt work for some reason
            if source_file.path == Path(file):
                for node in source_file.functions + source_file.classes:
                    if node.name == name and node.node_type.value == node_type:
                        return node

        raise Exception(f"Node not found : {name} in {file}")

    def get_test_funcs(self) -> List[Function]:
        return reduce(
            lambda x, y: x + y,
            [test_file.test_funcs() for test_file in self.test_files],
            [],
        )

    def get_test_classes(self) -> List[Class]:
        classes = reduce(
            lambda x, y: x + y,
            [test_file.test_classes() for test_file in self.test_files],
            [],
        )
        return classes

    def iter_tests(self) -> Iterator[Tuple[TestFile, Class]]:
        for test_file in self.test_files:
            for test_class in test_file.test_classes():
                yield test_file, test_class

    def find_file(self, file_path: str) -> Optional[SourceFile]:
        """
        Returns the file object
        """
        for file in self.source_files:
            # NOTE: need path here to deal with consistent / and \ in windows
            if Path(file.path) == Path(file_path):
                return file

    def write_file(self, file_path: str):
        """
        Writes the content of a SourceFile to its original file on disk.
        Note that this API is implemented on SourceRepo because we want
        to control all I/O interactions through this class
        """
        src_file = self.find_file(file_path)
        with open(self.repo_path / file_path, "w") as f:
            f.write(src_file.to_code())



from typing import List, Tuple
import asyncio

from .models import BaseModel

from .utils import (
    # TurboModel,
    extract_python_code,
    extract_yaml_code,
    extract_json_code,
)


async def invoke_llm_async(
    prompt: str,
    model: BaseModel,
    n_times: int,
    output: str = "str",
) -> List[str]:
    output = []

    coroutines = []
    for _ in range(n_times):
        coroutines.append(model.query(prompt))

    llm_outputs = await asyncio.gather(*coroutines)

    # should add the other methods here
    for out in llm_outputs:
        if output == "yaml":
            out = extract_yaml_code(out)
        elif output == "json":
            out = extract_json_code(out)
        elif output == "code":
            out = extract_python_code(out)

    return llm_outputs



from typing import List, Optional

from pydantic import BaseModel, validator
from cowboy.db.core import Database


class PythonConf(BaseModel):
    cov_folders: List[str]
    interp: str
    test_folder: Optional[str]
    pythonpath: Optional[str]

    @validator("interp")
    def validate_interp(cls, v):
        import os

        if not os.path.exists(v):
            raise ValueError(f"Interpreter path {v} does not exist")
        return v


class RepoConfig(BaseModel):
    repo_name: str  # of form owner_repo
    url: str
    cloned_folders: List[
        str
    ]  # list of cloned folders used for parallelizing run_test; many to many relationship
    # with instantiated repo contexts
    source_folder: (
        str  # source folder used for read/temp write operations; one to many relations
    )
    # pytest specific confs (although they could be generally applicable)
    python_conf: "PythonConf"
    is_experiment: Optional[bool] = False

    @validator("url")
    def validate_url(cls, v):
        import re

        if not re.match(r"^https:\/\/github\.com\/[\w-]+\/[\w-]+(\.git)?$", v):
            raise ValueError(
                "URL must be a valid GitHub HTTPS URL and may end with .git"
            )
        # if v.endswith(".git"):
        #     raise ValueError("URL should not end with .git")
        if re.match(r"^git@github\.com:[\w-]+\/[\w-]+\.git$", v):
            raise ValueError("SSH URL format is not allowed")
        return v

    def __post_init__(self):
        if isinstance(self.python_conf, dict):
            self.python_conf = PythonConf(**self.python_conf)

    def serialize(self):
        return {
            "repo_name": self.repo_name,
            "url": self.url,
            "cloned_folders": self.cloned_folders,
            "source_folder": self.source_folder,
            "python_conf": self.python_conf.__dict__,
            "is_experiment": self.is_experiment,
        }


class RepoConfigRepository:
    def __init__(self, db: Database):
        self.db = db

    def save(self, repo_config: RepoConfig):
        self.db.save_dict(
            dict_key="repos", key=repo_config.repo_name, value=repo_config.serialize()
        )

    def delete(self, repo_name: str):
        self.db.delete_dict("repos", repo_name)

    def find(self, repo_name: str) -> Optional[RepoConfig]:
        repo_config = self.db.get_dict("repos", repo_name)
        if repo_config:
            return self.rcfg_from_dict(repo_config)

        return None

    def rcfg_from_dict(self, d: dict) -> RepoConfig:
        python_conf = PythonConf(**d["python_conf"])
        return RepoConfig(
            repo_name=d["repo_name"],
            url=d["url"],
            cloned_folders=d.get("cloned_folders", []),
            source_folder=d.get("source_folder", []),
            python_conf=python_conf,
        )



from typing import List, Optional, Tuple
from pathlib import Path
from difflib import unified_diff

from cowboy_lib.ast import Class, Function, ASTNode, NodeType
from cowboy_lib.ast import PythonAST
from cowboy_lib.utils import locate_python_interpreter

from logging import getLogger
from copy import deepcopy

import subprocess


logger = getLogger("test_results")
longterm_logger = getLogger("longterm")


class LintException(Exception):
    pass


class SameNodeException(Exception):
    pass


class NodeNotFound(Exception):
    pass


class SourceFile:
    def __init__(
        self,
        lines: List[str],
        path: Path,
        language: str = "python",
    ):
        self._path = path
        self._lang = language

        self.functions: List[Function] = []
        self.classes: List[Class] = []
        self.lines: List[str] = []
        # self.imports: List[str] = []

        self.update_file_state(lines)

    def clone(self) -> "SourceFile":
        return deepcopy(self)

    @property
    def path(self):
        return self._path

    def update_file_state(self, lines: List[str]):
        """
        Updates instance variables
        """
        assert isinstance(lines, list)

        # REFACTOR-AST: start here first, this is where we convert
        # AST into our code-object representation of the src code
        self.ast_parser = PythonAST("\n".join(lines))
        self.functions, self.classes = self.ast_parser.parse()
        self.lines = lines

    def __repr__(self):
        return f"{self.path}"

    def diff(self, other: "SourceFile"):
        if not isinstance(other, SourceFile):
            raise TypeError("Can only diff SourceFile instances")

        a = self.to_code().splitlines(keepends=True)
        b = other.to_code().splitlines(keepends=True)
        diff = "".join(unified_diff(a, b))

        return diff

    # this would be another easy test for modification in test_parsing
    # commit: c58ded2
    def find_class(self, class_name: str) -> Optional[Class]:
        """
        Finds a class by name
        """
        return self.find_by_nodetype(class_name, node_type=NodeType.Class)

    def find_function(self, function_name: str) -> Optional[Function]:
        """
        Finds a function by name
        """
        return self.find_by_nodetype(function_name, node_type=NodeType.Function)

    def find_by_nodetype(
        self, node_name: str, node_type: NodeType = NodeType.Function
    ) -> Optional[List[ASTNode]]:
        """
        Finds a function or class by name
        """
        assert type(node_name) == str

        all = [f for f in self.functions + self.classes if f.name == node_name]
        filtered = [f for f in all if f.type == node_type]

        if len(filtered) > 1:
            raise SameNodeException(
                "More than one node found with the same name: ", node_name
            )
        elif len(filtered) == 0:
            raise NodeNotFound(
                "No node found with the given name, did you forget to put NODETYPE param again you dumbass?: ",
                node_name,
                node_type,
                self.path,
            )

        return filtered[0]

    def append(self, lines: str, class_name: Optional[str] = None) -> None:
        """
        Appends lines to the test file or to an existing class
        """
        if class_name:
            node = self.find_by_nodetype(class_name, node_type=NodeType.Class)
            _, end = node.range
        else:
            end = len(self.lines) - 1

        lines = lines.split("\n")
        lines = self.lines[: end + 1] + lines + self.lines[end + 1 :]
        self.update_file_state(self.to_linted_code(lines).split("\n"))

    def delete(self, node_name: str, node_type: NodeType = NodeType.Function) -> None:
        """
        Deletes a new function or class to the file, and updates SourceFile instance accordingly without hitting file
        """

        node = self.find_by_nodetype(node_name, node_type=node_type)
        start, end = node.range

        longterm_logger.info(f"Deleting: {node.name} => {start} to {end}")

        lines = self.lines[:start] + self.lines[end + 1 :]
        self.update_file_state(self.to_linted_code(lines).split("\n"))

    # TODO: specify other code linter here
    def to_linted_code(self, lines) -> str:
        """
        Lint generated code file
        """
        self.i = 0 if getattr(self, "i", None) is None else self.i + 1
        interp = locate_python_interpreter()

        black_cmd_str = f"{interp} -m black "
        tmp_file = f"/tmp/test{str(self.i)}.py"
        with open(tmp_file, mode="w+t", encoding="utf-8") as temp_file:
            # temp_file_name = temp_file.name
            temp_file.write("\n".join(lines))
            temp_file.flush()

            process = subprocess.Popen(
                black_cmd_str + tmp_file,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = process.communicate()
            if stderr:
                stderr = stderr.decode("utf-8")
                if "error:" in stderr:
                    raise LintException(f"Error while linting: {stderr}")

            with open(tmp_file, "r") as temp_file:
                linted_code = temp_file.read()

        return linted_code

    def map_line_to_node(
        self, start: int, end: int
    ) -> Optional[Tuple[ASTNode, ASTNode]]:
        """
        Finds the function and/or class that contains the line
        """
        for node in self.functions:
            if node.range[0] <= start and end <= node.range[1]:
                return node, node.scope
        return None, None

    def to_code(self) -> str:
        """
        Converts the sourcefile to code
        """

        return "\n".join(self.lines)

    def to_num_lines(self) -> str:
        return "\n".join([f"{i}: {line}" for i, line in enumerate(self.lines)])

    def to_llm_repr(self) -> str:
        repr_str = ""
        repr_str += "\n".join([f.__str__() for f in self.functions if not f.scope])
        repr_str += "\n"
        repr_str += "\n".join([c.__str__() for c in self.classes])
        return repr_str


class TestFile(SourceFile):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def test_nodes(self) -> List[ASTNode]:
        return [n for n in self.functions + self.classes if n.is_test]

    def test_funcs(self) -> List[Function]:
        return [func for func in self.functions if func.is_test]

    def test_classes(self) -> List[Class]:
        return [c for c in self.classes if c.is_test]

    def __repr__(self):
        return f"{self.path}"

    def new_test_funcs(self, new_file: "TestFile") -> List[Function]:
        """
        Returns a list of nodes that are in the new file but not in the current file
        """
        assert Path(self._path) == Path(new_file.path)

        return [
            f
            for f in new_file.test_funcs()
            if f.name not in [my_f.name for my_f in self.test_funcs()]
        ]



# These are API interactions initiated by the client, but have no corresponding CLI commands (in release)
from .build_mapping import api_baseline
from .augment import api_augment
from .register import api_register
from .get_tms import api_get_tms
from .remote_head import api_get_head



from pathlib import Path
import os
import click
import sys
import shutil
import subprocess
from git import Repo

from cowboy.db.core import Database
from cowboy.utils import gen_random_name

from cowboy.repo.models import RepoConfig
from cowboy.config import REPO_ROOT
from cowboy.exceptions import CowboyClientError

ALL_REPO_CONF = "src/config"
NUM_CLONES = 2


def del_file(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    import stat

    # Is the error an access error?
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


def create_cloned_folders(
    repo_conf: RepoConfig, repo_root: Path, db: Database, num_clones: int
):
    """
    Clones the repo from the forked_url
    """
    cloned_folders = []
    repo = db.get(repo_conf.repo_name)
    if repo:
        click.secho("Repo already exists", fg="red")
        sys.exit()

    clone_dir = repo_root / repo_conf.repo_name
    if clone_dir.exists():
        click.secho(
            f"Cloned folder {str(repo_root / repo_conf.repo_name)} already exist, skipping creating anew ...",
            fg="green",
        )
        return [str(f) for f in clone_dir.iterdir() if f.is_dir()]

    for i in range(num_clones):
        cloned_path = clone_repo(clone_dir, repo_conf.url)
        setuppy_init(repo_conf.repo_name, cloned_path, repo_conf.python_conf.interp)

        cloned_folders.append(str(cloned_path))

    return cloned_folders


def get_cloned_folders(repo_name: str):
    """
    Returns the cloned_folders for a repo
    """
    repo_folder = Path(REPO_ROOT) / repo_name
    if not repo_folder.exists():
        raise CowboyClientError(f"Repo {repo_name} does not exist")

    return [f for f in repo_folder.iterdir() if f.is_dir()]


# TODO: here there may be errors
def setuppy_init(repo_name: str, cloned_path: Path, interp: str):
    """
    Initialize setup.py file for each interpreter
    """
    cmd = ["cd", str(cloned_path), "&&", interp, "setup.py", "install"]

    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        text=True,
    )

    stdout, stderr = proc.communicate()
    # if stderr:
    #     print("Error while setup.py install")
    #     print(stderr)


def clone_repo(clone_dir: Path, repo_url: str) -> Path:
    """
    Creates a clone of the repo locally
    """
    dest_folder = clone_dir / gen_random_name()
    if dest_folder.exists():
        os.makedirs(dest_folder)

    Repo.clone_from(repo_url, dest_folder)

    return dest_folder


def delete_cloned_folders(repo_root: Path, repo_name: str):
    """
    Deletes cloned folders
    """
    try:
        repo_path = repo_root / repo_name
        shutil.rmtree(repo_path, onerror=del_file)
    except FileNotFoundError:
        return



import json
import os
from typing import Dict

from cowboy.exceptions import CowboyClientError
from cowboy.config import DB_PATH


class KeyNotFoundError(CowboyClientError):
    def __init__(self, key):
        super().__init__(f"Key {key} not found in DB")


class Database:
    """
    KV DB impl
    """

    _instance = None

    def __new__(cls, db_path: str = DB_PATH):
        if cls._instance is None:
            cls._instance = super(Database, cls).__new__(cls)
        return cls._instance

    def __init__(self, db_path: str = DB_PATH):
        """
        Initialize empty json at db_path
        """
        self.db_path = db_path

        if not os.path.exists(db_path):
            if os.path.isdir(db_path):
                os.makedirs(os.path.dirname(db_path), exist_ok=True)
            with open(db_path, "w") as f:
                f.write("{}")

        # in case emtpy file or db gets corrupted
        elif os.path.exists(db_path):
            try:
                with open(db_path, "r") as f:
                    json.load(f)
            except json.JSONDecodeError:
                print("DB corrupted, re-initializing ...")
                with open(db_path, "w") as f:
                    f.write("{}")

    def save_upsert(self, key, value):
        """
        Overwrites key if it exists, otherwise creates it
        """
        try:
            data = self.get_all()
            data[key] = value
            with open(self.db_path, "w") as f:
                json.dump(data, f)
        except IOError as e:
            print(f"Error saving to DB file: {e}")

    def save_dict(self, dict_key, key, value):
        """
        Adds a key/val pair to existing key
        """
        try:
            data = self.get_all()
            if not data.get(dict_key, None):
                data[dict_key] = {}

            data[dict_key][key] = value
            with open(self.db_path, "w") as f:
                json.dump(data, f, indent=2)

        except IOError as e:
            print(f"Error saving to DB file: {e}")

    def save_to_list(self, key, value):
        """
        Adds a value to a list
        """
        try:
            data = self.get_all()
            if not data.get(key, None):
                data[key] = []

            data[key].append(value)
            with open(self.db_path, "w") as f:
                json.dump(data, f, indent=2)

        except IOError as e:
            print(f"Error saving to DB file: {e}")

    def delete_from_list(self, key, value):
        """
        Deletes a value from a list
        """
        try:
            data = self.get_all()
            data[key].remove(value)
        except KeyError:
            return

        with open(self.db_path, "w") as f:
            json.dump(data, f)

    def get(self, key, default=None):
        return self.get_all().get(key, default)

    def get_dict(self, dict_key, key):
        data = self.get_all()
        if dict_key in data:
            return data[dict_key].get(key, None)
        return None

    def delete_dict(self, dict_key, key):
        data = self.get_all()
        del data[dict_key][key]
        with open(self.db_path, "w") as f:
            json.dump(data, f)

    def delete(self, key):
        data = self.get_all()
        if key in data:
            del data[key]
            with open(self.db_path, "w") as f:
                json.dump(data, f)
        else:
            raise KeyNotFoundError(key)

    def reset(self):
        with open(self.db_path, "w") as f:
            f.write("{}")

    def get_all(self) -> Dict:
        try:
            if os.path.exists(self.db_path):
                with open(self.db_path, "r") as f:
                    return json.load(f)
            else:
                return {}
        except IOError as e:
            print(f"Error reading from DB file: {e}")
            return {}



from cowboy.config import API_ENDPOINT
from cowboy.db.core import Database
from cowboy.exceptions import CowboyClientError
from cowboy.utils import start_daemon

from urllib.parse import urljoin
import requests
import logging
import json

logger = logging.getLogger(__name__)


def parse_pydantic_error(error_json):
    """
    Parse the Pydantic error JSON object and format it into a readable string message.

    :param error_json: JSON object returned by Pydantic containing error details.
    :return: Formatted string message.
    """
    error_obj = json.loads(error_json)
    details = error_obj.get("detail", [])

    messages = []
    for detail in details:
        loc = detail.get("loc", [])
        msg = detail.get("msg", "")
        error_type = detail.get("type", "")

        location = " -> ".join(map(str, loc))
        messages.append(f"Location: {location}\nMessage: {msg}\nType: {error_type}\n")

    return "\n".join(messages)


class HTTPError(Exception):
    pass


class InternalServerError(Exception):
    pass


class APIClient:
    def __init__(self, db: Database):
        self.server = API_ENDPOINT
        self.db = db

        # secondary auth token for our janky poll auth
        self.pauth_token = None

        # polling state
        self.encountered_401s = 0

    def headers(self):
        token = self.db.get("token", "")
        headers = {"Authorization": f"Bearer {token}"}

        return headers

    def get(self, uri: str):
        url = urljoin(self.server, uri)

        res = requests.get(url, headers=self.headers())

        return self.parse_response(res)

    def post(self, uri: str, data: dict):
        url = urljoin(self.server, uri)
        res = requests.post(url, json=data, headers=self.headers())

        return self.parse_response(res)

    def poll(self):
        """
        Polls the server for new tasks that comes through. Reason we implement
        this method differently than others is because we require some pretty
        janky logic -> basically an alternative auth token
        """
        headers = self.headers()
        if self.pauth_token:
            headers.update({"x-task-auth": self.pauth_token})

        url = urljoin(self.server, "/task/get")
        res = requests.get(url, headers=headers)

        task_token = res.headers.get("set-x-task-auth", None)
        if task_token:
            self.pauth_token = task_token

        # next two conds are used to detect when the server restarts
        if self.pauth_token and res.status_code == 401:
            self.encountered_401s += 1

        if self.encountered_401s > 3:
            self.pauth_token = None
            self.encountered_401s = 0

        return res.json()

    def long_post(self, uri: str, data: dict):
        """
        Need this method to handle long requests, because requests consume
        all sigints while waiting for the response to return, we have to wrap
        it in a new thread and use is_alive/join(timeout) to allow the sigint
        to reach the main thread
        """
        return start_daemon(self.post, (uri, data))

    def long_get(self, uri: str):
        """
        Need this method to handle long requests, because requests consume
        all sigints while waiting for the response to return, we have to wrap
        it in a new thread and use is_alive/join(timeout) to allow the sigint
        to reach the main thread
        """
        return start_daemon(self.get, (uri,))

    def delete(self, uri: str):
        url = urljoin(self.server, uri)

        res = requests.delete(url, headers=self.headers())

        return self.parse_response(res)

    def parse_response(self, res: requests.Response):
        """
        Parses token from response and handles HTTP exceptions, including retries and timeouts
        """
        if res.status_code == 401:
            raise HTTPError("Unauthorized, are you registered or logged in?")

        elif res.status_code == 500:
            raise InternalServerError()

        elif res.status_code == 400 or res.status_code == 422:
            message = res.json()["detail"]

            if isinstance(message, (dict, list)):
                message = json.dumps(message, indent=2)

            raise CowboyClientError(message)

        return res.json()



from cowboy_lib.repo import GitRepo

from cowboy.repo.models import RepoConfig, RepoConfigRepository, PythonConf
from cowboy.repo.repo import (
    create_cloned_folders,
    delete_cloned_folders,
    get_cloned_folders,
)
from cowboy.api_cmds import (
    api_baseline,
    api_augment,
    api_register,
    api_get_tms,
    api_get_head,
)
from cowboy.task_client import Manager
from cowboy.browser import serve_ui
from cowboy.exceptions import CowboyClientError
from cowboy import config

from cowboy.db.core import Database
from cowboy.db.public import init_react_env_vars
from cowboy.http import APIClient, InternalServerError

import click
import yaml
from pathlib import Path
import json

db = Database()
api = APIClient(db)
rc_repo = RepoConfigRepository(db)


def owner_name_from_url(url: str):
    owner, repo_name = url.split("/")[-2:]
    return owner, repo_name


@click.group()
def cowboy_cli():
    """Command-line interface to Cowboy."""
    pass


@cowboy_cli.command("dump")
def dump():
    """Dumps db.json for debugging."""
    print(db.get_all())


@cowboy_cli.group("user")
def cowboy_user():
    """Container for all user commands."""
    pass


# TODO: make initialization user dialogue instead
@cowboy_user.command("init")
def init():
    """Initializes user account for Cowboy."""
    try:
        with open(".user", "r") as f:
            user_conf = yaml.safe_load(f)
    except FileNotFoundError:
        click.secho('User definition file ".user" does not exist', fg="red")
        return

    # only allow one user to be registered at a time
    registered = db.get("registered", False)
    if registered:
        click.secho(
            "We are currently only supporting one user per client. If you want to re-register, "
            f"or something has gone wrong, manually delete {config.DB_PATH} and run cowboy user init again.",
            fg="red",
        )
        return

    token = api_register(user_conf)

    # save token in python db
    db.save_upsert("token", token)
    db.save_upsert("registered", True)
    db.save_upsert("user", user_conf["email"])

    # initialize the front-end config after
    init_react_env_vars(token, config.API_ENDPOINT)

    click.secho(
        "Successfully registered user. You can delete .user in case you dont want "
        "passwords/sensitive tokens to be exposed locally",
        fg="green",
    )


@cowboy_user.command("update-oai-key")
@click.argument("openai_api_key")
def update_oai(openai_api_key):
    """Updates the openapi key for the user."""

    api.post("/user/update/openai-key", {"openai_api_key": openai_api_key})
    click.secho("Successfully updated openapi key", fg="green")


@cowboy_user.command("reset")
def reset():
    """Resets user account for Cowboy ."""
    api.long_get(f"/user/delete")

    for repo in db.get("repos", []):
        delete_cloned_folders(Path(config.REPO_ROOT), repo)

    db.reset()
    click.secho("Successfully reset user data", fg="green")


@cowboy_cli.group("repo")
def cowboy_repo():
    """Container for all repo commands."""
    pass


@cowboy_repo.command("create")
@click.argument("config_path")
def repo_init(config_path):
    """Initializes a new repo."""
    try:
        with open(config_path, "r") as f:
            repo_config = yaml.safe_load(f)
    except FileNotFoundError:
        click.secho("Config file does not exist.", fg="red")
        return

    repo_name = repo_config["repo_name"]
    click.echo("Initializing new repo {}".format(repo_name))

    python_conf = PythonConf(
        cov_folders=repo_config.get("cov_folders", []),
        test_folder=repo_config.get("test_folder", ""),
        interp=repo_config.get("interp"),
        pythonpath=repo_config.get("pythonpath", ""),
    )

    repo_config = RepoConfig(
        repo_name=repo_name,
        url=repo_config.get("url"),
        cloned_folders=[],
        source_folder="",
        python_conf=python_conf,
        is_experiment=repo_config.get("is_experiment", False),
    )

    db.save_to_list("repos", repo_name)
    cloned_folders = create_cloned_folders(
        repo_config, Path(config.REPO_ROOT), db, config.NUM_CLONES
    )
    repo_config.cloned_folders = cloned_folders

    try:
        api.post("/repo/create", repo_config.serialize())
        print(json.dumps(repo_config.serialize(), indent=4))
        click.secho("Successfully created repo: {}".format(repo_name), fg="green")

    except Exception as e:
        error_message = str(e)
        click.secho(f"Repo creation failed on server: {error_message}", fg="red")
        click.secho(f"Rolling back repo creation", fg="red")

        db.delete_from_list("repos", repo_name)

        # leave cloned folders alone for debugging TestSuiteError
        if "RunnerError" in error_message:
            return

        delete_cloned_folders(Path(config.REPO_ROOT), repo_name)
        return


@cowboy_repo.command("delete")
@click.argument("repo_name")
def delete(repo_name):
    """
    Deletes all repos and reset the database
    """
    try:
        api.delete(f"/repo/delete/{repo_name}")
    except Exception:
        click.secho(f"Failed to delete repo {repo_name}", fg="red")
        return

    db.delete_from_list("repos", repo_name)
    delete_cloned_folders(Path(config.REPO_ROOT), repo_name)
    click.secho(f"Deleted repo {repo_name}", fg="green")


NO_SUPPORT_REPO_UPDATES = """
A thousand apologies, dear and valued user, our current backend does not track the HEAD of your remote 
repo; we are currently working on this feature. Will still proceed with the command 
"""

SIT_TIGHT = """
Sit tight this might take a while (>5min) 
"""


@cowboy_repo.command("augment")
@click.argument("repo_name")
@click.option("--mode", required=False)
# NOTE: currently not supported for now
# @click.option("--files", required=False, multiple=True)
@click.option("--tm", required=False, multiple=True)
def augment(repo_name, mode, tm):
    """
    Augments existing test modules with new test cases
    """
    # for grammars sake..
    tms = tm

    if tms:
        mode = "module"
    elif mode == "auto" and tms:
        raise Exception("Cannot specify file or tms when mode=auto")

    repo_folder = get_cloned_folders(repo_name)[0]
    remote_head = api_get_head(repo_name)
    local_head = GitRepo(repo_folder).local_commit
    if remote_head != local_head:
        click.secho(
            f"Local repo HEAD {local_head[:7]} is not up-to-date with remote {remote_head[:7]}\n"
            + NO_SUPPORT_REPO_UPDATES,
            fg="yellow",
        )

    click.secho(SIT_TIGHT, fg="yellow")
    # NOTE: might need to expose baseline command manually to user
    api_baseline(repo_name, mode, tms)
    session_id = api_augment(repo_name, mode, tms)
    serve_ui(session_id)


@cowboy_repo.command("baseline")
@click.argument("repo_name")
@click.option("--mode", required=False)
# NOTE: currently not supported for now
# @click.option("--files", required=False, multiple=True)
@click.option("--tm", required=False, multiple=True)
def augment(repo_name, mode, tm):
    """
    Augments existing test modules with new test cases
    """
    # for grammars sake..
    tms = tm

    if tms:
        mode = "module"
    elif mode == "auto" and tms:
        raise Exception("Cannot specify file or tms when mode=auto")

    # NOTE: might need to expose baseline command manually to user
    api_baseline(repo_name, mode, tms)


@cowboy_repo.command("get_tms")
@click.argument("repo_name")
def get_tms(repo_name):
    """
    Gets all test_modules for a given repo
    """
    tms = api_get_tms(repo_name)
    for tm in tms:
        print("TestModule: ", tm["name"])
        print("TestFile: ", tm["filepath"])
        for i, unit_test in enumerate(tm["unit_tests"], start=1):
            print(f"{i}. {unit_test}")

        print("\n")


@cowboy_cli.command("browser")
@click.argument("session_id")
def browser(session_id):
    serve_ui(session_id)


def entrypoint():
    """The entry that the CLI is executed from"""

    try:
        import cowboy.logger

        # TODO: we should make a note that currently only supporting
        # running a single repo-at-a-time usage, due to hb and error file conflicts
        runner = Manager(config.HB_PATH, config.HB_INTERVAL)
        cowboy_cli()
    except CowboyClientError as e:
        click.secho(
            f"CowboyClientError: {e}",
            bold=True,
            fg="red",
        )
    except Exception as e:
        raise e
        error_msg = f"ERROR: {e}"
        if db.get("debug", False):
            import traceback

            tb = traceback.format_exc()
            error_msg = f"ERROR: {e}\n{tb}"

        click.secho(error_msg, bold=True, fg="red")


if __name__ == "__main__":
    entrypoint()



from cowboy_lib.utils import find_substring

from typing import List, Tuple, Optional, NewType, NamedTuple
from pathlib import Path
import re
from enum import Enum

SHA = NewType("SHA", str)


class NoTestCtxtException(Exception):
    pass


class HunkChunk:
    """
    A hunk in a diff
    """

    def __init__(self, body: str):
        self.body = body

        lines = self.body.split("\n")
        o_start, o_end, n_start, n_end = self._parse_hunk_header(lines[0])
        self.old_range = (o_start, o_end)
        self.new_range = (n_start, n_end)

        self.plus_blob, self.minus_blob = self._parse_hunk_lines(body)
        self.new_func = self._new_func_decl()

    def _parse_hunk_lines(self, body: str) -> Tuple[List[str], List[str]]:
        """
        Parses a hunk in a diff
        """
        plus_lines = []
        minus_lines = []
        for line in body.split("\n"):
            if line.startswith("+"):
                plus_lines.append(line[1:])
            elif line.startswith("-"):
                minus_lines.append(line[1:])

        return "\n".join(plus_lines), "\n".join(minus_lines)

    def _parse_hunk_header(self, line: str) -> Optional[Tuple[int, int]]:
        """
        Returns the line number of the hunk
        """
        pattern = r"@@ -(\d+),?(\d+)? \+(\d+),?(\d+)? @@"

        # Use regex to find matches in the hunk
        match = re.search(pattern, line)

        def convert_to_int(matched):
            if matched:
                return int(matched)
            if matched is None:
                return -1

        if match:
            old_start, old_count, new_start, new_count = map(
                convert_to_int, match.groups()
            )
            return old_start, old_start + old_count, new_start, new_start + new_count

        return None, None, None, None

    def _new_func_decl(self) -> str:
        """
        If the hunk declares a new test function, return it
        """
        # Hunk can be empty since we are only currently looking at + lines
        # if len(self.lines) == 0:
        #     return ""

        PYTHON_FUN_DEF = r"\+?\s*(?:async\s+)?def\s+([a-zA-Z_][a-zA-Z_0-9]*)\s*\("
        match = re.search(PYTHON_FUN_DEF, self.body)
        if match:
            return match.group(1)
        else:
            return ""


class DiffMode(Enum):
    MODIFIED = "modified"
    DELETED = "deleted"
    NEW = "new"
    UNKNOWN = "unknown"


class DiffAttr(NamedTuple):
    a_path: str
    b_path: Optional[str]
    b_path_fallback: str  # have no idea wtf this thing is
    mode: DiffMode


class Diff:
    def __init__(self, body: str):
        self.body = body
        self.attrs: DiffAttr = self._parse_patch_attributes()
        self.hunks: List[HunkChunk] = self._parse_hunks()
        # lets assume this is the most reliable path, because it *should* account
        # for new/renamed files
        # remove b/ from begining
        self.filepath: str = self._find_new_filepath()

    def _find_new_filepath(self) -> str:
        """
        Finds the filepath of the b/file
        """
        if self.attrs:
            return self.norm_path(self.attrs.b_path_fallback)

        # look for the +++ line
        for line in self.body.split("\n"):
            if line.startswith("+++"):
                return line[3:].strip()

    def norm_path(self, path: str) -> str:
        """
        Removes the "a/" and "b/" prefixes
        """
        return path[2:]

    def _parse_hunks(self) -> List[HunkChunk]:
        hunks = []
        hunk_starts = find_substring(self.body, "@@")

        try:
            start = hunk_starts.pop(0)
        except IndexError:
            return []

        for end in hunk_starts:
            hunk_body = self.body[start:end]
            start = end
            hunks.append(HunkChunk(hunk_body))
        hunks.append(HunkChunk(self.body[start:]))  # last hunk

        return hunks

    def _parse_patch_attributes(self) -> DiffAttr:
        """
        Stolen from GitPython (properly attribute later)
        """
        re_header = re.compile(
            rb"""
                                    ^diff[ ]--git
                                        [ ](?P<a_path_fallback>"?[ab]/.+?"?)[ ](?P<b_path_fallback>"?[ab]/.+?"?)\n
                                    (?:^old[ ]mode[ ](?P<old_mode>\d+)\n
                                    ^new[ ]mode[ ](?P<new_mode>\d+)(?:\n|$))?
                                    (?:^similarity[ ]index[ ]\d+%\n
                                    ^rename[ ]from[ ](?P<rename_from>.*)\n
                                    ^rename[ ]to[ ](?P<rename_to>.*)(?:\n|$))?
                                    (?:^new[ ]file[ ]mode[ ](?P<new_file_mode>.+)(?:\n|$))?
                                    (?:^deleted[ ]file[ ]mode[ ](?P<deleted_file_mode>.+)(?:\n|$))?
                                    (?:^similarity[ ]index[ ]\d+%\n
                                    ^copy[ ]from[ ].*\n
                                    ^copy[ ]to[ ](?P<copied_file_name>.*)(?:\n|$))?
                                    (?:^index[ ](?P<a_blob_id>[0-9A-Fa-f]+)
                                        \.\.(?P<b_blob_id>[0-9A-Fa-f]+)[ ]?(?P<b_mode>.+)?(?:\n|$))?
                                    (?:^---[ ](?P<a_path>[^\t\n\r\f\v]*)[\t\r\f\v]*(?:\n|$))?
                                    (?:^\+\+\+[ ](?P<b_path>[^\t\n\r\f\v]*)[\t\r\f\v]*(?:\n|$))?
                                """,
            re.VERBOSE | re.MULTILINE,
        )

        diff_attrs = []

        try:
            # Encode normally, expecting no surrogates
            encoded_body = bytes(self.body, encoding="utf-8")
        except UnicodeEncodeError:
            # Handle the presence of surrogates explicitly
            encoded_body = bytes(self.body, encoding="utf-8", errors="surrogateescape")

        for _header in re_header.finditer(encoded_body):
            (
                a_path_fallback,
                b_path_fallback,
                old_mode,
                new_mode,
                rename_from,
                rename_to,
                new_file_mode,
                deleted_file_mode,
                copied_file_name,
                a_blob_id,
                b_blob_id,
                b_mode,
                a_path,
                b_path,
            ) = _header.groups()

            a_path = self.norm_path(a_path) if a_path else b""
            b_path = self.norm_path(b_path) if b_path else b""
            file_mode = None

            if new_file_mode:
                file_mode = DiffMode.NEW
            elif deleted_file_mode:
                file_mode = DiffMode.DELETED
            elif a_path and b_path and a_path == b_path:
                file_mode = DiffMode.MODIFIED
            else:
                file_mode = DiffMode.UNKNOWN

            diff_attrs.append(
                DiffAttr(
                    a_path.decode("utf-8"),
                    b_path.decode("utf-8"),
                    b_path_fallback.decode("utf-8") if b_path_fallback else "",
                    file_mode,
                )
            )
        # at this point we've already split the diff into sections
        return diff_attrs[0] if len(diff_attrs) > 0 else []

    def __str__(self):
        return self.body


class DiffsNotFoundException(Exception):
    pass


# this class needs a rewrite ...
# TODO: add file mode to each diff chunk
class CommitDiff:
    """
    Class to represent a diff/patch

    FYI patch is a diff that can be applied directly to a filepath
    """

    # TODO: Too much logic in the __init__ method, move inside methods instead
    def __init__(
        self,
        patch: str,
        timestamp: str = None,
    ):
        if not patch:
            raise DiffsNotFoundException("Empty patch file argument")
        try:
            self.diffs = self._segment_diffs(patch)
            self.files = [d.filepath for d in self.diffs]
            self.hunks = [hunk for d in self.diffs for hunk in d.hunks]
            self._timestamp = timestamp
        except Exception as e:
            print(f"Patch parsing error:\n{e}")
            print(f"Patch:\n{patch}")

    def code_diffs(self):
        return [diff for diff in self.diffs if self.is_code_file(diff.filepath)]

    def is_code_file(self, filename: str):
        return "test" not in filename and filename.endswith(".py")

    def is_test_file(self, filename: str):
        return "test" in filename and filename.endswith(".py")

    @property
    def timestamp(self) -> str:
        return getattr(self, "_timestamp", None)

    def find_diff(self, filename: str) -> Optional[Diff]:
        """
        Returns a diff corresponding to the filename
        """
        for diff in self.diffs:
            if Path(diff.filepath) == Path(filename):
                return diff
        return None

    # @property
    # def code_diff(self) -> str:
    #     """
    #     All diffs pertaining to non-test case files joined together
    #     """
    #     diff = "\n".join(
    #         [diff.body for diff in self.diffs if diff.filepath not in self.test_files]
    #     )
    #     if diff:
    #         return diff + "\n"
    #     # not sure
    #     else:
    #         return ""

    # @property
    # def tests_diff(self) -> str:
    #     """
    #     All diffs pertaining to pytest unit test cases joined together
    #     """
    #     diff = "\n".join(
    #         [diff.body for diff in self.diffs if diff.filepath in self.test_files]
    #     )
    #     if diff:
    #         return diff + "\n"
    #     else:
    #         return ""

    # def __str__(self):
    #     return self.code_diff + "\n" + self.tests_diff

    @property
    def test_files(self) -> List[str]:
        return [d.filepath for d in self.diffs if self.is_test_file(d.filepath)]

    @property
    def code_files(self) -> List[str]:
        return [d.filepath for d in self.diffs if self.is_code_file(d.filepath)]

    def _segment_diffs(self, patch: str) -> List[Diff]:
        """
        Segments the patch into diffs, where each diff marks whether a file has been
        modified, created anew, or deleted
        """

        diff_sections = []
        diff_starts = find_substring(patch, "diff --git")
        if not diff_starts:
            diff_starts = find_substring(patch, "---")

        try:
            start = diff_starts.pop(0)
        except IndexError:
            raise DiffsNotFoundException("No diff sections found in patch")

        for end in diff_starts:
            diff_sections.append(patch[start:end])
            start = end
        diff_sections.append(patch[start:])

        try:
            return [Diff(lines) for lines in diff_sections]
        except Exception as e:
            import traceback

            print(f"Error parsing diff:\n{e}")
            traceback.print_exc()



import logging

# from anthropic import Anthropic, HUMAN_PROMPT, AI_PROMPT
from dataclasses import dataclass, fields
from openai import BadRequestError, OpenAI, AzureOpenAI
from simple_parsing.helpers import FrozenSerializable, Serializable
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_not_exception_type,
)

# from typing import Optional

from openai import AsyncOpenAI

logger = logging.getLogger("test_results")


@dataclass(frozen=True)
class ModelArguments(FrozenSerializable):
    model_name: str
    api_key: str
    per_instance_cost_limit: float = 0.0
    total_cost_limit: float = 0.0
    temperature: float = 1.0
    top_p: float = 1.0
    replay_path: str = None
    host_url: str = "localhost:11434"


@dataclass
class APIStats(Serializable):
    total_cost: float = 0
    instance_cost: float = 0
    tokens_sent: int = 0
    tokens_received: int = 0
    api_calls: int = 0

    def __add__(self, other):
        if not isinstance(other, APIStats):
            raise TypeError("Can only add APIStats with APIStats")

        return APIStats(
            **{
                field.name: getattr(self, field.name) + getattr(other, field.name)
                for field in fields(self)
            }
        )

    def replace(self, other):
        if not isinstance(other, APIStats):
            raise TypeError("Can only replace APIStats with APIStats")

        return APIStats(
            **{field.name: getattr(other, field.name) for field in fields(self)}
        )


class ContextWindowExceededError(Exception):
    pass


class CostLimitExceededError(Exception):
    pass


class BaseModel:
    MODELS = {}
    SHORTCUTS = {}

    def __init__(self, args: ModelArguments):
        self.args = args
        self.model_metadata = {}
        self.stats = APIStats()

        # Map `model_name` to API-compatible name `api_model`
        self.api_model = (
            self.SHORTCUTS[self.args.model_name]
            if self.args.model_name in self.SHORTCUTS
            else self.args.model_name
        )

        # Map model name to metadata (cost, context info)
        MODELS = {
            **{dest: self.MODELS[src] for dest, src in self.SHORTCUTS.items()},
            **self.MODELS,
        }
        if args.model_name in MODELS:
            self.model_metadata = MODELS[args.model_name]
        else:
            raise ValueError(
                f"Unregistered model ({args.model_name}). Add model name to MODELS metadata to {self.__class__}"
            )

    def reset_stats(self, other: APIStats = None):
        if other is None:
            self.stats = APIStats(total_cost=self.stats.total_cost)
            logger.info("Resetting model stats")
        else:
            self.stats = other

    def update_stats(self, input_tokens, output_tokens):
        """
        Calculates the cost of a response from the openai API.

        Args:
        input_tokens (int): The number of tokens in the prompt.
        output_tokens (int): The number of tokens in the response.

        Returns:
        float: The cost of the response.
        """
        # Calculate cost and update cost related fields
        cost = (
            self.model_metadata["cost_per_input_token"] * input_tokens
            + self.model_metadata["cost_per_output_token"] * output_tokens
        )
        self.stats.total_cost += cost
        self.stats.instance_cost += cost
        self.stats.tokens_sent += input_tokens
        self.stats.tokens_received += output_tokens
        self.stats.api_calls += 1

        # Log updated cost values to std. out.
        logger.info(
            f"input_tokens={input_tokens:_}, "
            f"output_tokens={output_tokens:_}, "
            f"instance_cost={self.stats.instance_cost:.2f}, "
            f"cost={cost:.2f}"
        )
        logger.info(
            f"total_tokens_sent={self.stats.tokens_sent:_}, "
            f"total_tokens_received={self.stats.tokens_received:_}, "
            f"total_cost={self.stats.total_cost:.2f}, "
            f"total_api_calls={self.stats.api_calls:_}"
        )

        # Check whether total cost or instance cost limits have been exceeded
        # if (
        #     self.args.total_cost_limit > 0
        #     and self.stats.total_cost >= self.args.total_cost_limit
        # ):
        #     logger.warning(
        #         f"Cost {self.stats.total_cost:.2f} exceeds limit {self.args.total_cost_limit:.2f}"
        #     )
        #     raise CostLimitExceededError("Total cost limit exceeded")

        # if (
        #     self.args.per_instance_cost_limit > 0
        #     and self.stats.instance_cost >= self.args.per_instance_cost_limit
        # ):
        #     logger.warning(
        #         f"Cost {self.stats.instance_cost:.2f} exceeds limit {self.args.per_instance_cost_limit:.2f}"
        #     )
        #     raise CostLimitExceededError("Instance cost limit exceeded")
        return cost

    async def query(self, prompt: str) -> str:
        raise NotImplementedError("Use a subclass of BaseModel")


class OpenAIModel(BaseModel):
    MODELS = {
        "gpt-3.5-turbo-0125": {
            "max_context": 16_385,
            "cost_per_input_token": 5e-07,
            "cost_per_output_token": 1.5e-06,
        },
        "gpt-3.5-turbo-1106": {
            "max_context": 16_385,
            "cost_per_input_token": 1.5e-06,
            "cost_per_output_token": 2e-06,
        },
        "gpt-3.5-turbo-16k-0613": {
            "max_context": 16_385,
            "cost_per_input_token": 1.5e-06,
            "cost_per_output_token": 2e-06,
        },
        "gpt-4-32k-0613": {
            "max_context": 32_768,
            "cost_per_input_token": 6e-05,
            "cost_per_output_token": 0.00012,
        },
        "gpt-4-0613": {
            "max_context": 8_192,
            "cost_per_input_token": 3e-05,
            "cost_per_output_token": 6e-05,
        },
        "gpt-4-1106-preview": {
            "max_context": 128_000,
            "cost_per_input_token": 1e-05,
            "cost_per_output_token": 3e-05,
        },
        "gpt-4-0125-preview": {
            "max_context": 128_000,
            "cost_per_input_token": 1e-05,
            "cost_per_output_token": 3e-05,
        },
    }

    SHORTCUTS = {
        "gpt3": "gpt-3.5-turbo-1106",
        "gpt3-legacy": "gpt-3.5-turbo-16k-0613",
        "gpt4": "gpt-4-1106-preview",
        "gpt4-legacy": "gpt-4-0613",
        "gpt4-0125": "gpt-4-0125-preview",
        "gpt3-0125": "gpt-3.5-turbo-0125",
    }

    def __init__(self, args: ModelArguments):
        super().__init__(args)

        self.client = AsyncOpenAI(api_key=args.api_key)

    def history_to_messages(
        self, history: list[dict[str, str]], is_demonstration: bool = False
    ) -> list[dict[str, str]]:
        """
        Create `messages` by filtering out all keys except for role/content per `history` turn
        """
        # Remove system messages if it is a demonstration
        if is_demonstration:
            history = [entry for entry in history if entry["role"] != "system"]
            return "\n".join([entry["content"] for entry in history])
        # Return history components with just role, content fields
        return [
            {k: v for k, v in entry.items() if k in ["role", "content"]}
            for entry in history
        ]

    @retry(
        wait=wait_random_exponential(min=1, max=15),
        reraise=True,
        stop=stop_after_attempt(3),
        retry=retry_if_not_exception_type((CostLimitExceededError, RuntimeError)),
    )
    async def query(self, prompt: str) -> str:
        """
        Query the OpenAI API with the given `history` and return the response.
        """

        print("Querying MDOEL ....")
        try:
            # Perform OpenAI API call
            response = await self.client.chat.completions.create(
                messages=[{"role": "user", "content": prompt}],
                model=self.api_model,
            )

            input_tokens = response.usage.prompt_tokens
            output_tokens = response.usage.completion_tokens
            self.update_stats(input_tokens, output_tokens)
            return response.choices[0].message.content

        except BadRequestError as e:
            raise CostLimitExceededError(
                f"Context window ({self.model_metadata['max_context']} tokens) exceeded"
            )



import re
from typing import Tuple, List

from dataclasses import dataclass


@dataclass
class LMModelSpec:
    model: str
    cost: float
    ctxt_window: int


def extract_yaml_code(llm_output: str) -> str:
    try:
        return re.search(r"```yaml\n(.*)```", llm_output, re.DOTALL).group(1)
    except AttributeError:
        return llm_output


def extract_python_code(llm_output: str) -> str:
    try:
        return re.search(r"```python\n(.*)```", llm_output, re.DOTALL).group(1)
    except AttributeError:
        return llm_output


def extract_json_code(llm_output: str) -> str:
    try:
        return re.search(r"```json\n(.*)```", llm_output, re.DOTALL).group(1)
    except AttributeError:
        return llm_output


def extract_test_functions(llm_output: str) -> List[str]:
    PYTHON_FUN_DEF = r"\+?\s*(?:async\s+)?def\s+([a-zA-Z_][a-zA-Z_0-9]*)\s*\("

    return re.findall(PYTHON_FUN_DEF, llm_output)



from cowboy.http import APIClient
from cowboy.db.core import Database


db = Database()
api = APIClient(db)


def api_get_head(repo_name):
    """
    Gets all test_modules for a given repo
    """
    commit = api.get(f"/repo/get_head/{repo_name}")

    return commit["sha"]



from typing import NamedTuple, Optional, List, Dict, Iterable, Tuple
from pathlib import Path
from itertools import product
from typing import List, Tuple, NewType, Dict
import re


from logging import getLogger

logger = getLogger("test_results")


class CoverageException(Exception):
    pass


class CoverageFailure(Exception):
    pass


class CoverageSubtractionError(Exception):
    def __init__(self):
        super().__init__("Negative covered but contributed covered_lines")


def get_full_path(base_path: Path, filename: str):
    """
    Join filename, which is relative to the source repo root to the base repo path, while
    also converting it to use the same Path format (WindowsPath or PosixPath)
    """
    filename = filename.replace("\\", "/")
    return Path(base_path / filename)


class Coverage:
    def __init__(
        self,
        filename: str,
        covered_lines: List[int],
        missing_lines: List[int],
    ):
        self.all_lines = covered_lines + missing_lines
        assert len(self.all_lines) == len(set(self.all_lines))

        # self.cov: int = cov
        self.covered_lines: List[int] = covered_lines
        self.missing_lines: List[int] = missing_lines

        self.stmts: int = len(self.all_lines)
        self.misses: int = len(self.missing_lines)
        self.covered: int = len(self.covered_lines)

        if covered_lines and missing_lines:
            for line in covered_lines:
                if line in missing_lines:
                    raise CoverageException(
                        f"Line {line} is both covered and missing in {filename}"
                    )

        self.filename = self.convert_path(filename)

        # defer initialization
        self._covered_lines_dict: Dict[int, str] = {}
        self._miss_lines_dict: Dict[int, str] = {}

    def convert_path(self, filename: str):
        """
        Converts the path to a Unix-style path
        """
        return filename.replace("\\", "/")

    @property
    def cov(self) -> float:
        return self.covered / self.stmts

    def __eq__(self, other: Optional["Coverage"]):
        # None comparison
        if not other:
            if (
                self.filename == None
                and self.stmts == None
                and self.misses == None
                and self.covered == None
                and self.cov == None
            ):
                return True
            return False

        elif isinstance(other, Coverage):
            if (
                self.filename == other.filename
                and self.stmts == other.stmts
                and self.misses == other.misses
                and self.covered == other.covered
                and self.cov == other.cov
            ):
                return True
            return False

        else:
            raise CoverageException("Comparisons only allowed for Coverage or None")

    def __sub__(self, other: "Coverage"):
        try:
            assert isinstance(other, Coverage)
            assert self.filename == other.filename
            # may be from another commit
            assert self.stmts == other.stmts
        except AssertionError as e:
            raise CoverageException(f"Assertion failed: {e}")

        # NOTE:
        # use set sub here to find only the overlapping missing lines
        # for eg.
        # a_miss = [1,2,3]
        # b_miss = [1,2,4]
        # if we just sub the abs len of missing lines, we would get:
        # a_miss - b_miss = []
        # but instead we want:
        # a_miss - b_miss = [3]
        # Because we want to if b improved the coverage of a
        added_lines = set(self.covered_lines) - set(other.covered_lines)
        missing_lines = set(self.all_lines) - added_lines

        cov = Coverage(
            self.filename,
            covered_lines=list(added_lines),
            missing_lines=list(missing_lines),
        )
        # need to do this to support negative coverage
        cov.covered = self.covered - other.covered
        if cov.covered < 0 and cov.covered_lines:
            # theoretically shud not happen, except in case
            # where a covered test causes another test to fail
            raise CoverageSubtractionError

        return cov

    def __add__(self, other: "Coverage"):
        """
        Unlike sub, we can only add the covered from other if it does not overlap
        with a pre-exsiting line
        """
        try:
            assert isinstance(other, Coverage)
            assert self.filename == other.filename
            # may be from another commit
            assert self.stmts == other.stmts
        except AssertionError as e:
            raise CoverageException(f"Assertion failed: {e}")

        added_lines = set(other.covered_lines) - set(self.covered_lines)
        missing_lines = set(self.all_lines) - set(self.covered_lines) - added_lines

        return Coverage(
            self.filename,
            covered_lines=list(set(self.covered_lines).union(added_lines)),
            missing_lines=list(missing_lines),
        )

    @classmethod
    def diff_cov(cls, a: "Coverage", b: "Coverage", keep_line: int) -> "Coverage":
        """
        Gets the diff between two coverages that also includes covered_lines
        """
        assert keep_line in [1, 2]

        cov_diff = a - b

        sub1 = set(a.covered_lines) if keep_line == 1 else set(b.covered_lines)
        sub2 = set(b.covered_lines) if keep_line == 1 else set(a.covered_lines)
        if len(sub1) < len(sub2):
            logger.warn("a < b in the subtraction of covered lines, is this expected?")

        covered_lines = sub1 - sub2
        cov_diff.covered_lines = covered_lines
        return cov_diff

    def __str__(self):

        return f"Coverage: {self.filename}, stmts: {self.stmts}, misses: {self.misses}, covered: {self.covered}"

    def read_line_contents(self, base_path: Path):
        """
        Lazily reads the line contents of the file
        """
        fp = get_full_path(base_path, self.filename)

        with open(fp, "r", encoding="utf-8") as file:
            # print("Covered lines: ", self.covered_lines)
            all_lines = ""
            for i, line in enumerate(file.read().split("\n"), start=1):
                all_lines += f"{i}. {line}" + "\n"
                if i in self.covered_lines:
                    self._covered_lines_dict[i] = line
                elif i in self.missing_lines:
                    self._miss_lines_dict[i] = line

    def print_lines(self, line_type: str = "covered"):
        lines_dict = (
            self._covered_lines_dict
            if line_type == "covered"
            else self._miss_lines_dict
        )

        repr = ""
        repr += f"{line_type} lines in :: {self.filename}\n"
        for k, v in lines_dict.items():
            repr += f"{k}: {v}\n"

        return repr

    def get_contiguous_lines(self) -> Iterable[List[Tuple[int, str]]]:
        """
        Returns a list of contiguous line groups
        """
        from itertools import groupby

        for k, g in groupby(
            enumerate(self._covered_lines_dict.items()), lambda ix: ix[1][0] - ix[0]
        ):
            yield [x for _, x in g]

    def serialize(self):
        return {
            "filename": self.filename,
            "covered_lines": self.covered_lines,
            "missing_lines": self.missing_lines,
        }

    @classmethod
    def deserialize(self, data) -> "Coverage":
        return Coverage(data["filename"], data["covered_lines"], data["missing_lines"])


class NoCoverageDB(Exception):
    pass


class TestCoverage:
    """
    Coverage for a list of files from a commit
    """

    def __init__(
        self,
        cov_list: List[Coverage],
        isdiff: bool = False,
    ):
        self.isdiff = isdiff
        self.filenames = [cov.filename for cov in cov_list]

        self._cov_list = cov_list

        total_misses = 0
        total_stmts = 0
        total_covered = 0
        for coverage in cov_list:
            total_misses += coverage.misses
            total_stmts += coverage.stmts
            total_covered += coverage.covered

        self.total_cov = Coverage("TOTAL", [], [])
        self.total_cov.misses = total_misses
        self.total_cov.stmts = total_stmts
        self.total_cov.covered = total_covered

    @property
    def cov_list(self):
        return [cov for cov in self._cov_list if cov.filename != "TOTAL"]

    # REFACTOR-RUNNER: re-implement as a mixin-method on TestCoverage
    @classmethod
    def from_coverage_file(cls, coverage_json: dict) -> "TestCoverage":
        cov_list = []

        if not coverage_json:
            return cls(cov_list)

        # add exception catcher here for missing json
        for filename, data in coverage_json["files"].items():
            cov_list.append(
                Coverage(
                    filename,
                    data["executed_lines"],
                    data["missing_lines"],
                )
            )

        return cls(cov_list)

    @classmethod
    def diff_cov(
        cls, a: "TestCoverage", b: "TestCoverage", keep_line: int
    ) -> "TestCoverage":
        """
        Used for subtracting two TestCoverages when we also want to diff their covered_lines
        keep_line parameter is introduced to control the order of the set subtraction
        """
        if keep_line not in [1, 2]:
            raise ValueError("keep_line must be 1 or 2")

        cov_list = []
        for a, b in zip(a.cov_list, b.cov_list):
            cov_diff = Coverage.diff_cov(a, b, keep_line)
            cov_list.append(cov_diff)

        return cls(cov_list, isdiff=True)

    def get_file_cov(self, filename: Path, base_path: Path) -> Optional[Coverage]:
        """
        Gets a file coverage by filename. Base_path required because coverage file paths are relative
        to the repo_path as root
        """
        try:
            return next(
                filter(
                    lambda x: (
                        x.filename
                        if not base_path
                        else base_path / x.filename == filename
                    ),
                    self.cov_list,
                )
            )
        except StopIteration:
            return None

    def __bool__(self):
        return self.cov_list != []

    def __iter__(self):
        return iter([cov for cov in self.cov_list if cov.filename != "TOTAL"])

    def is_zero(self):
        return self.total_cov.misses == 0

    def __sub__(self, other: "TestCoverage") -> "TestCoverage":
        """
        Take the difference of every matching file coverage plus our own coverage
        that is not matched by other and that is nonzero
        """
        a, b = self.cov_list, other.cov_list

        merged_list = []
        a_only_cov = [a for a in a if a.filename not in [b.filename for b in b]]
        intersect_fp = [i.filename for i in a if i.filename in [b.filename for b in b]]

        for cov_a, cov_b in product(a, b):
            if cov_a.filename == cov_b.filename and cov_a.filename in intersect_fp:
                cov_diff = cov_a - cov_b
                if cov_diff.covered != 0:
                    merged_list.append(cov_diff)

        return TestCoverage(merged_list + a_only_cov, isdiff=True)

    def __add__(self, other: "TestCoverage") -> "TestCoverage":
        """
        Take add to our existing coverage only those lines that are not covered by ours
        """
        a, b = self.cov_list, other.cov_list

        merged_list = []
        a_only_cov = [a for a in a if a.filename not in [b.filename for b in b]]
        b_only_cov = [b for b in b if b.filename not in [a.filename for a in a]]
        intersect_fp = [i.filename for i in a if i.filename in [b.filename for b in b]]

        for cov_a, cov_b in product(a, b):
            if cov_a.filename == cov_b.filename and cov_a.filename in intersect_fp:
                merged_list.append(cov_a + cov_b)

        return TestCoverage(a_only_cov + merged_list + b_only_cov, isdiff=True)

    def covered_lines(self) -> List[Tuple[str, List[int]]]:
        cov_lines = [(cov.filename, cov.covered_lines) for cov in self.cov_list]
        return cov_lines

    def missing_lines(self) -> List[Tuple[str, List[int]]]:
        cov_lines = [(cov.filename, cov.missing_lines) for cov in self.cov_list]
        return cov_lines

    def cov_list_str(self):
        return "\n".join([str(cov) for cov in self.cov_list])

    def __repr__(self):
        return f"TestCoverage: {self.total_cov}, IsDiff:{self.isdiff}"

    def serialize(self):
        return {
            "cov_list": [cov.serialize() for cov in self.cov_list],
            "isdiff": self.isdiff,
        }

    @classmethod
    def deserialize(self, data: List[List]) -> "TestCoverage":
        return TestCoverage(
            [Coverage(**lines) for lines in data["cov_list"]],
            isdiff=data["isdiff"],
        )


TestError = NewType("TestError", str)


class CoverageResult:
    """
    Represents the result of a coverage run
    """

    def __init__(self, stdout: str, stderr: str, coverage_json: Dict):
        self.coverage: TestCoverage = TestCoverage.from_coverage_file(coverage_json)
        # self.coverage2 = TestCoverage.from_coverage_report(stdout)

        self.total_tests = self._parse_total_tests(stdout)
        self.failed, self.total_failed = self._parse_failed_tests(stdout)
        self.stderr = stderr
        # generated functions
        self.gen_funcs = []

    def to_dict(self) -> Dict:
        return {"coverage": self.coverage.serialize(), "failed": self.failed}

    def _parse_total_tests(self, stdout: str):
        """
        Parse total number of tests from pytest output
        """
        pattern = r"collected\s+(\d+)\s+items"
        match = re.search(pattern, stdout)
        if match:
            return int(match.group(1))
        return 0

    # TODO: parse pytest errors as well as failure
    def _parse_failed_tests(
        self, stdout: str
    ) -> Tuple[List[Tuple[str, TestError]], int]:
        """
        Parse every failed test from pytest output
        """
        pattern = r"FAILED\s+(?:\S+?)::(\S+?)\s+-"
        failed_modules = re.findall(pattern, stdout)

        # NOTE: currently treating parameterized tests as single tests
        total_failed = set()

        # parse test_module names
        for failed_test in failed_modules:
            # logger.info(f"Failed tests: {failed_test}")

            if "[" in failed_test:
                failed_test = failed_test.split("[")[0]

            if "::" in failed_test:
                test_module = failed_test.split("::")[0]
                failed_test = failed_test.split("::")[1]
                total_failed.add(f"{test_module}.{failed_test}")

            total_failed.add(failed_test)

        logger.info(f"Total failed tests: {len(failed_modules)}")

        # parse error info
        pattern = r"_{2,}(\s+\b[\w\.]+)(?:\[\S+\])?\s+_{2,}\n(.*?)\n[_|-]"
        test_info = re.findall(pattern, stdout, re.DOTALL)

        return {f.strip(): error.rstrip() for f, error in test_info}, len(total_failed)

    def get_failed(self, test_name):
        """
        Did test_name fail in this coverage run?
        """
        return self.failed.get(test_name, None)

    def __bool__(self):
        return bool(self.coverage)

    def get_coverage(self):
        return self.coverage

    # actually parse out the stderr
    def get_error(self):
        if not self.stderr:
            raise Exception("No error found")
        return self.stderr



from cowboy.runner.python import PytestDiffRunner
from cowboy_lib.api.runner.shared import RunTestTaskArgs
from cowboy.http import APIClient
from cowboy.db.core import Database
from cowboy.repo.models import RepoConfig

import threading
from concurrent.futures import ThreadPoolExecutor

if __name__ == "__main__":
    db = Database()
    api = APIClient(db)
    repo_config = RepoConfig(**api.get(f"/repo/get/test3"))

    # Initialize PytestDiffRunner
    runner = PytestDiffRunner(repo_config)

    # Setup RunTestTaskArgs
    task_args = RunTestTaskArgs(patch_file=None, exclude_tests=[], include_tests=[])

    round = 0
    while True:
        from datetime import datetime

        current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print("STARTING ROUND: ", round, "AT TIME: ", current_time)

        def run_fake_test():
            print("Running fake test suite")
            runner.fake_test_suite(task_args)

        with ThreadPoolExecutor(max_workers=5) as executor:
            futures = [executor.submit(run_fake_test) for _ in range(5)]
            for future in futures:
                future.result()

        round += 1



from cowboy_lib.coverage import Coverage, get_full_path
from cowboy_lib.repo.source_file import TestFile, ASTNode, NodeNotFound, NodeType
from cowboy_lib.utils import get_current_git_commit

from typing import List, Optional, TYPE_CHECKING, Tuple
from pathlib import Path

from cowboy_lib.repo.source_repo import SourceRepo
from cowboy_lib.test_modules.target_code import TargetCode


class IncompatibleCommit(Exception):
    pass


class TestModule:
    """
    A TestFile with a list of selected Tests
    """

    def __init__(
        self,
        test_file: TestFile,
        nodes: List[ASTNode],
        commit_sha: str,
        # hack: to check if we are not desync'd from the repo
        check_commit: bool = False,
        chunks: List[TargetCode] = [],
    ):
        test_folder = Path(*test_file.path.parts[:2])
        if check_commit:
            assert commit_sha == get_current_git_commit(test_folder)

        self.commit_sha = commit_sha
        self.test_file = test_file
        self.chunks: List[TargetCode] = chunks
        self.cov_diff = None

        # class or test file
        self.nodes = nodes
        self._isclass = True if self.nodes[0].type is NodeType.Class else False
        self.name = (
            self.test_file.path.name if not self._isclass else self.nodes[0].name
        )

    def __eq__(self, other: "TestModule"):
        return self.test_file.path + self.name == other.test_file.path + other.name

    @property
    def tests(self) -> List[ASTNode]:
        return (
            self.nodes
            if not self._isclass
            else [
                func
                for func in self.test_file.test_funcs()
                if func.scope.name == self.name
            ]
        )

    def targeted_files(self):
        """
        The list of files that the test module is testing
        """
        # kinda of a hack, but we always assume that base path is in the form
        # repos/<repo_name> appended to the filepath
        # fp = lambda p: base_path if base_path else Path(*p.parts[2:])

        return list(set([c.filepath for c in self.chunks]))

    @property
    def path(self):
        return self.test_file.path

    def target_test_file(self) -> Path:
        """
        Returns the test file that will be modified by test strategy
        to generate new tests
        """
        return self.test_file.path

    def num_lines(self, filename: Optional[str] = ""):
        if not filename:
            chunks = self.chunks
        else:
            chunks = [c for c in self.chunks if c.filepath == filename]

        return sum(
            [c.range[1] - c.range[0] if c.range[1] - c.range[0] else 1 for c in chunks]
        )

    def get_test_code(self, curr_commit: str):
        """
        We are purposely including curr_commit here so TestFile will stayed sync'd with fs
        Lazily loads the test file just in case we want to setup GitRepo to match
        commit_sha. Pass curr_commit as a parameter to double check
        """
        if curr_commit != self.commit_sha:
            raise IncompatibleCommit(
                f"curr_commit: {curr_commit} does not match {self.commit_sha}"
            )

        return "\n\n".join([test.to_code() for test in self.tests])

    def did_change(self, new_file: TestFile):
        """
        Compares the test file ranges
        """
        for n in new_file.test_nodes:
            for test_node in self.tests:
                if (
                    n.name == test_node.name
                    and n.range != test_node.range
                    # NOTE: this takes into account line additions and deletions
                    # but not modifications, we leaving that for later when the diff
                    # parsing interface gets built
                    and n.range[1] - n.range[0]
                    != test_node.range[1] - test_node.range[0]
                ):
                    new_mods = n.range[1] - n.range[0]
                    old_test = test_node.range[1] - test_node.range[0]
                    if new_mods > old_test:
                        print("Tests were added!")
                    else:
                        print("Tests were deleted!")
                    return True
        return False

    def set_chunks(
        self,
        changed_coverage: List[Coverage],
        source_repo: "SourceRepo",
        base_path: Path = None,
    ):
        """
        Gets the missing/covered lines of each of the coverage differences
        """
        self.chunks = []
        for cov in changed_coverage:
            if cov.filename == "TOTAL":
                raise Exception("TOTAL COVERAGE FILE FOUND")

            cov.read_line_contents(base_path)
            for l_group in cov.get_contiguous_lines():
                start = l_group[0][0]
                end = l_group[-1][0]
                range = (start, end)

                src_file = source_repo.find_file(cov.filename)
                func, cls = src_file.map_line_to_node(start, end)

                lines = [g[1] for g in l_group]

                print("Setting chunk with filepath: ", str(cov.filename))

                chunk = TargetCode(
                    range=range,
                    lines=lines,
                    # could also just move the logic into TestModuleMixin
                    filepath=str(cov.filename),
                    func_scope=func if func else "",
                    class_scope=cls if cls else "",
                )
                self.chunks.append(chunk)

    def delete_chunk(self, range, filepath):
        for c in self.chunks[:]:
            if c.range == range and c.filepath == filepath:
                self.chunks.remove(c)

    def diff_chunks(self, other: "TestModule"):
        return [c for c in self.chunks if c not in other.chunks]

    def print_chunks(self) -> str:
        if not self.chunks:
            return ""

        repr = ""
        repr += f"Test Module: {self.test_file.path}\n"
        curr_file = self.chunks[0].filepath.name
        repr += f"File: {self.chunks[0].filepath.name}\n"

        for c in self.chunks:
            if c.filepath.name != curr_file:
                curr_file = c.filepath.name
                repr += f"File: {c.filepath.name}\n"
            repr += c.to_lines()
        return repr



from abc import ABC, abstractmethod
from cowboy_lib.api.runner.shared import RunTestTaskArgs
from cowboy_lib.coverage import CoverageResult

from typing import Tuple, List, Any


class Runner(ABC):
    """
    Runs the lang/framework specific unit test
    """

    @abstractmethod
    def run_test(self, args: RunTestTaskArgs) -> Tuple[CoverageResult, str, str]:
        """
        Run the test and return the coverage result
        """
        raise NotImplementedError

    def _construct_cmd(
        self,
        repo_path,
        selected_args_str: str = "",
        deselected_args_str: str = "",
    ):
        """
        Constructs the cmd for running the test via subprocess
        """
        raise NotImplementedError

    def _get_include_tests_arg_str(self, include_tests: List[str] = []):
        """
        Constructs the arg string for selecting specific tests
        """
        raise NotImplementedError

    def _get_exclude_tests_arg_str(self, exclude_tests: List[Any]):
        """
        Constructs the arg string for excluding specific tests
        """
        raise NotImplementedError



from cowboy.config import TASK_ENDPOINT
from cowboy.runner import PytestDiffRunner, RunnerError
from cowboy.db.core import Database
from cowboy.repo.models import RepoConfig
from cowboy.http import APIClient
from cowboy_lib.api.runner.shared import RunTestTaskArgs, Task, TaskResult, TaskType

from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from pathlib import Path
import threading
import _thread
import time
import json
from requests import ConnectionError
import traceback

from logging import getLogger

log = getLogger(__name__)


class BGClient:
    """
    Single Task client that runs as a subprocess in the background
    and fetches tasks from server
    """

    def __init__(
        self,
        api: APIClient,
        fetch_endpoint: str,
        heart_beat_fp: Path,
        heart_beat_interval: int = 5,
        sleep_interval=5,
    ):
        self.api = api
        self.db = db
        self.run_executor = ThreadPoolExecutor(max_workers=5)
        self.fetch_endpoint = fetch_endpoint

        # curr tasks : technically dont need since we await every new
        # tasks via runner.acquire_one() but use for debugging
        self.curr_t = []
        self.completed = 0

        self.runners = {}

        # retrieved tasks
        self.lock = threading.Lock()
        self.retrieved_t = []
        self.start_t = []

        # heartbeat
        self.heart_beat_fp = Path(heart_beat_fp)
        self.heart_beat_interval = heart_beat_interval

        # run tasks
        t1 = threading.Thread(target=self.start_heartbeat, daemon=True)
        t2 = threading.Thread(target=self.start_polling, daemon=True)

        t1.start()
        t2.start()

    def get_runner(self, repo_name: str) -> PytestDiffRunner:
        """
        Initialize or retrieve ean existing runner for Repo
        """
        # technically we can store all info locally but dont feel like
        # managing repo state de-sync
        repo_conf = self.api.get(f"/repo/get/{repo_name}")

        # need to lock access to self.runners since two threads that
        # instantiate the same runner will not clobber the test repository
        # when they try to execute tests in the same folder without
        # LockedRepos in PytestDiffRunner
        with self.lock:
            if self.runners.get(repo_name, None):
                return self.runners[repo_name]

            repo_conf = RepoConfig(**repo_conf)
            runner = PytestDiffRunner(repo_conf)
            self.runners[repo_name] = runner

        return runner

    def start_polling(self):
        """
        Polls server and receive tasks. Currently only two types:'cowboy.repo.runner'
        1. Run Test -> runs in separate thread
        2. Shutdown -> shutdown client immediately
        """
        while True:
            try:
                task_res = self.api.poll()
                if task_res:
                    log.info(f"Receieved {len(task_res)} tasks from server")
                    for t in task_res:
                        task_type = t["type"]
                        log.info(f"Received task: {t}")
                        if task_type == TaskType.RUN_TEST:
                            task = Task(**t)
                            task.task_args = RunTestTaskArgs.from_json(**t["task_args"])
                            self.curr_t.append(task.task_id)

                            threading.Thread(
                                target=self.run_test_thread, args=(task,)
                            ).start()

                        elif task_type == TaskType.SHUTDOWN:
                            log.info(f"Received shutdown signal")
                            self.complete_task(Task(**t))
                            # sends sigint to main thread
                            _thread.interrupt_main()

            except ConnectionError as e:
                # log.error("Error connecting to server ...")
                pass

            # These errors result from how we handle server restarts
            # and our janky non-db auth method so can just ignore
            except Exception as e:
                log.error(
                    f"Exception from client: {e} : {type(e).__name__}\n{traceback.format_exc()}"
                )

            time.sleep(1.0)  # Poll every 'interval' second

    def run_test_thread(self, task: Task):
        """
        Runs task and updates its result field when finished
        """
        try:
            log.info(f"Starting task: {task.task_id}")

            runner = self.get_runner(task.task_args.repo_name)

            log.info(f"Running testsuite for {task.task_args.repo_name} ...")
            cov_res, *_ = runner.run_testsuite(task.task_args)
            task.result = TaskResult(**cov_res.to_dict())
            self.complete_task(task)

        except RunnerError as e:
            task.result = TaskResult(exception=str(e))
            self.complete_task(task)

            log.error(f"Exception from runner: {e}\n{traceback.format_exc()}")

    def complete_task(self, task: Task):
        self.api.post(f"/task/complete", json.loads(task.json()))

    def heart_beat(self):
        new_file_mode = False
        if not self.heart_beat_fp.exists():
            with open(self.heart_beat_fp, "w") as f:
                f.write("")

        with open(self.heart_beat_fp, "r") as f:
            raw = f.read()
            if len(raw) > 10**6:
                new_file_mode = True

        with open(self.heart_beat_fp, "w" if new_file_mode else "a") as f:
            curr_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            f.write(curr_time + "\n")

    def start_heartbeat(self):
        while True:
            threading.Thread(target=self.heart_beat, daemon=True).start()

            time.sleep(self.heart_beat_interval)


if __name__ == "__main__":
    import sys
    import logging
    from cowboy.logger import formatter

    def get_console_handler():
        """
        Returns a console handler for logging.
        """
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        return console_handler

    if len(sys.argv) < 3:
        log.info(
            "Usage: python client.py <heartbeat_file> <heartbeat_interval> <console>"
        )
        sys.exit(1)

    hb_path = sys.argv[1]
    hb_interval = int(sys.argv[2])
    console = bool(sys.argv[3])

    if console:
        log.addHandler(get_console_handler())

    db = Database()
    api = APIClient(db)
    BGClient(api, TASK_ENDPOINT, hb_path, hb_interval)
    # keep main thread alive so we can terminate all threads via sys interrupt
    # (because main thread is the only one we can send signals to)
    while True:
        time.sleep(1.0)



from cowboy.http import APIClient
from cowboy.db.core import Database

db = Database()
api = APIClient(db)


def api_register(user_conf):
    res = api.post(f"/user/register", user_conf)
    return res["token"]



from typing import List, Optional, Tuple, NewType, Union
from dataclasses import dataclass, field
from enum import Enum

import ast


@dataclass
class Argument:
    name: str
    type: Optional[str]

    def __str__(self):
        name = self.name
        type = f"{':' + self.type if self.type else ''}"
        return name + type


class ASTNode:
    def __init__(
        self,
        name: str,
        range: Tuple[int, int],
        scope: Optional["ASTNode"],
        decorators: List["Decorator"],
        lines: List[str],
        ast_node: Optional[ast.AST],
        is_test: bool = False,
        node_type: Optional["NodeType"] = None,
    ):
        # need to turn on mypy for this shit ...
        if decorators:
            assert isinstance(decorators[0], Decorator)
        else:
            assert isinstance(decorators, list) and len(decorators) == 0

        self._name = name
        # REFACTOR-AST: create a lang specific AST node class to account for decorators?
        # ie. PyAST/GolangAST
        self.decorators = decorators
        self.range = self._set_range(range)
        self.lines = [l.rstrip() for l in lines]
        self.is_test = is_test
        self.scope = scope
        self.ast_node = ast_node
        self.node_type = self.get_node_type()

    def get_node_type(self):
        return NodeType(self.__class__.__name__)

    def is_ast(self, other_ast: ast.AST) -> bool:
        """
        Checks if self matches another ast.AST node
        """
        return self.ast_node == other_ast

    def __eq__(self, other: "ASTNode"):
        return self.name + str(self.range) == other.name + str(other.range)

    def __hash__(self):
        return sum([ord(c) for c in self.name + str(self.range)])

    def _set_range(self, range) -> Tuple[int, int]:
        start = self.decorators[0].range[0] if self.decorators else range[0]
        end = range[1]
        return (start, end)

    def set_is_test(self, is_test: bool):
        self.is_test = is_test

    def to_code(self):
        repr = ""
        for dec in self.decorators:
            repr += dec.to_code()

        # newline if we have decorators
        repr += "\n" if repr else ""
        repr += "\n".join([l.rstrip() for l in self.lines])
        return repr

    @property
    def type(self) -> "NodeType":
        return NodeType(self.__class__.__name__)

    @property
    def name(self):
        raise NotImplementedError


@dataclass
class Decorator(ASTNode):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Class(ASTNode):
    def __init__(self, *args, **kwargs):
        self.functions: List[Function] = kwargs.pop("functions", [])

        super().__init__(*args, **kwargs)

    def add_func(self, func: "Function"):
        self.functions.append(func)

    # def __str__(self):
    #     funcs_str = "\n".join([f"{func.__str__()}" for func in self.functions])
    #     return f"Class: {self._name} \n{funcs_str}"

    def __eq__(self, other: "Class"):
        return self._name == other._name

    # def __repr__(self):
    #     funcs_str = "\n".join([f"{func}" for func in self.functions])
    #     return f"Class: {self._name} \n{funcs_str}"

    @property
    def name(self):
        return self._name


class Function(ASTNode):
    def __init__(self, *args, **kwargs):
        # should replace this with ... a prototype that contains .name property
        # self.scope: Class = kwargs.pop("scope", [])
        self.arguments = kwargs.pop("arguments", [])

        super().__init__(*args, **kwargs)

        self.is_test = True if self._name.startswith("test") else False

    def is_meth(self):
        return bool(self.scope)

    def __str__(self):
        return f"{self._name}({', '.join([arg.__str__() for arg in self.arguments])})"

    def is_method(self):
        return self.scope is not None

    def func_name(self):
        return self._name.split(".")[-1]

    @property
    def name(self):
        scope_prefix = f"{self.scope.name}." if self.scope else ""
        return f"{scope_prefix}{self._name}"


class NodeType(Enum):
    Function = Function.__name__
    Class = Class.__name__
    Decorator = Decorator.__name__



from cowboy.http import APIClient
from cowboy.db.core import Database


db = Database()
api = APIClient(db)


def api_baseline(repo_name, mode, tms):
    """
    Builds the test module to source file mapping for each selected
    test module
    """
    api.long_post(
        "/tm/build-mapping",
        {
            "repo_name": repo_name,
            "mode": mode,
            "tms": tms,
            "files": [],
            # TODO: change this to False
            "overwrite": True,
        },
    )



from cowboy.http import APIClient
from cowboy.db.core import Database


db = Database()
api = APIClient(db)


def api_get_tms(repo_name):
    """
    Gets all test_modules for a given repo
    """
    tms = api.get(f"/tm/{repo_name}")
    return tms



import importlib.resources as pkg_resources
from platformdirs import user_data_path
from pathlib import Path


CLIENT_MODE = "debug"

if CLIENT_MODE == "debug":
    # assume that we are executing from the root of the cowboy directory in debug mode
    COWBOY_DIR = Path(".")
elif CLIENT_MODE == "release":
    COWBOY_DIR = Path(user_data_path())

# Paths
REPO_ROOT = COWBOY_DIR / "repos"
USER_CONFIG = COWBOY_DIR / ".user"
DB_PATH = COWBOY_DIR / "db.json"
LOG_DIR = COWBOY_DIR / "logs"
COWBOY_FRONTEND_CONFIG = COWBOY_DIR / "build/config.json"

# Task client vars
REACT_CONFIG_PATH = (
    Path(str(pkg_resources.files("cowboy-client"))) / "build/config.json"
    if CLIENT_MODE == "release"
    else "build/config.json"
)
HB_PATH = COWBOY_DIR / ".heartbeat"
HB_INTERVAL = 2
NUM_CLONES = 3
RUN_TEST_TIMEOUT = 300

API_ENDPOINT = "http://18.223.150.134:3000"
TASK_ENDPOINT = f"{API_ENDPOINT}/task/get"



from cowboy_lib.utils import gen_random_name
from cowboy_lib.repo.diff import CommitDiff

import os
import tempfile
import hashlib
from pathlib import Path
from logging import getLogger
from git import Repo, GitCommandError
from dataclasses import dataclass
import shutil
import random
from typing import List, Union, Tuple
from pydantic import BaseModel

log = getLogger(__name__)


class NoRemoteException(Exception):
    pass


class NoMainBranch(Exception):
    pass


def del_file(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    import stat

    # Is the error an access error?
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


class GitRepo:
    """
    Used to manage git operations on a git repo
    """

    def __init__(self, repo_path: Path, remote: str = "origin", main: str = ""):
        if not repo_path.exists():
            # test suite may be renamed or deleted
            raise Exception("GitRepo does not exist: ", repo_path)

        # used for reversing patches
        self.patched_files = {}
        self.repo_folder = repo_path
        self.repo = Repo(repo_path)
        self.head = self.repo.head

        if main:
            if not self.branch_exists(main):
                raise NoMainBranch(main)
        else:
            potential = ["master", "main"]
            for branch in potential:
                main = branch if self.branch_exists(branch) else ""
                if main:
                    break
            if not main:
                raise NoMainBranch(main)

        self.main = main
        try:
            self.origin = self.repo.remotes.__getattr__(remote)
        except AttributeError:
            raise NoRemoteException(remote)

        self.username = self.origin.url.split("/")[-2]
        self.repo_name = self.origin.url.split("/")[-1]

        self.branch_prefix = "cowboy_"

    @classmethod
    def clone_repo(cls, clone_dst: Path, url: str) -> Path:
        """
        Creates a clone of the repo locally
        """
        if not os.path.exists(clone_dst):
            os.makedirs(clone_dst)  # Ensure the destination folder exists

        Repo.clone_from(url, clone_dst)
        return cls(clone_dst)

    @classmethod
    def delete_repo(cls, repo_dst: Path):
        """
        Deletes a repo from the db and all its cloned folders
        """
        import platform

        if not repo_dst.exists():
            return

        if platform.system() == "Windows":
            shutil.rmtree(repo_dst, onerror=del_file)
        else:
            shutil.rmtree(repo_dst)

    def reset_to_commit(self, commit_sha, parent=None, head: int = 0):
        """
        Resets the index of the repository to a specific commit.
        """
        self.repo.git.reset(f"--hard", commit_sha)
        return f"Successfully reset to commit {commit_sha}"

    def reset_to_commit_head(self, head: int = 0):
        head = f"HEAD~{str(head)}" if head else ""

        self.repo.git.reset(f"--hard", head)
        return f"Successfully reset to commit {head}"

    def commit_exists(self, commit_sha: str) -> bool:
        """
        Checks if a commit exists in the repo
        """
        try:
            self.repo.commit(commit_sha)
            return True
        except Exception:
            return False

    def get_curr_commit(self):
        """
        Returns the current commit sha
        """
        return self.head.commit.hexsha

    def get_prev_commit(self, commit_sha):
        """
        Returns the previous commit of a given commit sha
        """
        return self.repo.commit(commit_sha).parents[0]

    def apply_patch(self, patch: str) -> None:
        """
        Applies a patch from a .diff file to a single file in the repository
        """
        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as patch_file:
            patch_file.write(patch.encode("utf-8") + b"\n")
            patch_file.flush()

        patch_hash = hashlib.md5(patch.encode("utf-8")).hexdigest()
        self.repo.git.apply(patch_file.name, whitespace="nowarn")
        self.patched_files[patch_hash] = patch_file.name

    def reverse_patch(self, patch: str) -> None:
        """
        Reverses a patch from a .diff
        """
        patch_hash = hashlib.md5(patch.encode()).hexdigest()
        patch_file = self.patched_files[patch_hash]

        self.repo.git.apply(patch_file, reverse=True)
        self.patched_files.pop(patch_hash)

    def branch_exists(self, branch: str):
        """
        Checks if a branch exists in the repo
        """
        # print(f"Branch: {branch}")
        # print(f"Branches {[str(br) for br in self.repo.heads]} in {self.repo_folder}")
        if branch in [str(br) for br in self.repo.heads]:
            print(
                branch,
                [str(br) for br in self.repo.heads],
                branch in [str(br) for br in self.repo.heads],
            )
            return True

        return False

    def checkout(self, branch_name: str, new=False):
        """
        Checks out an existing branch in the repo
        """
        if new:
            branch = self.repo.create_head(branch_name)
        else:
            branch = self.repo.heads[branch_name]

        print(f"Checking out: {branch}")
        branch.checkout()

    def clean_branches(self, branch_prefix: str):
        """
        Deletes all branches with a specific prefix
        """
        removed = []
        for branch in self.repo.branches:
            if branch.name.startswith(branch_prefix):
                removed.append(branch.name)
                self.repo.delete_head(branch)

        return removed

    @property
    def remote_commit(self) -> str:
        """
        Gets the sha of latest commit on origin
        """
        self.repo.remotes.origin.fetch()
        remote_sha = self.repo.remotes.origin.refs.__getattr__(self.main).commit.hexsha

        return remote_sha

    @property
    def local_commit(self) -> str:
        return self.repo.head.commit.hexsha

    # TODO: move this into sync_repo
    def fetch_diffs(self) -> Tuple[CommitDiff, str]:
        """
        Diffs the remote with our local repo
        """
        try:
            if self.local_commit == self.remote_commit:
                print("No updates available.")
                return None, None
            else:
                print("Updates found!")

                # Get the diff between the old commit and the new HEAD
                diff = self.repo.git.diff(self.local_commit, self.remote_commit)
                commit_diff = CommitDiff(diff)

                return commit_diff, self.remote_commit
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def pull(self):
        self.repo.remotes.origin.pull()

    def add_n_commit(self, files: List[str], msg: str):
        self.repo.index.add(files)
        self.repo.index.commit(msg)

    def push(self, branch_name: str = "", force: bool = False):
        if not branch_name:
            branch_name = self.main
        self.origin.push(refspec=f"{branch_name}:{branch_name}", force=True)

    def checkout_and_push(
        self,
        name: str,
        commit_message: str,
        files_to_commit: list,
    ):
        """
        Checks out a new branch, commits changes, and pushes to the remote. Returns the
        URL for the merge request of our new branch against main

        Args:
        - name: The "suggested" name
        - commit_message: The commit message to use.
        - files_to_commit: List of file paths (relative to the repo root) to commit.

        Returns:
        - None
        """
        branch_name = name
        if self.branch_exists(name):
            branch_name = self.branch_prefix + name + "_" + gen_random_name()

        # Check out a new branch
        try:
            new_branch = self.repo.create_head(branch_name)
            new_branch.checkout()

            # Add and commit changes
            self.repo.index.add(files_to_commit)
            self.repo.index.commit(commit_message)
            self.push(branch_name=branch_name)
            origin_url = self.origin.url.replace(".git", "")
        except Exception as e:
            log.error(f"Exception in {self.repo_name}: {str(e)}")
            pass
        finally:
            self.checkout(self.main)
            log.info(f"Resetting to branch {self.main}")

        # url for branch merge request
        return f"{origin_url}/compare/{self.main}...{self.username}:{self.repo_name}:{branch_name}?expand=1"


class PatchApplyExcepion(Exception):
    pass


class IncompatibleCommit(Exception):
    pass


class PatchFile(BaseModel):
    path: Path
    patch: str


class PatchFileContext:
    """
    Context manager for applying and reversing patches
    """

    def __init__(
        self, repo: GitRepo, patch: Union[str, PatchFile], revert: bool = True
    ):

        self.repo = repo
        self.patch = patch
        # assume all cases repo and patch are both specified, or neither are
        self.head_commit = self.repo.head.commit if self.patch else None
        self.failed_id = random.randint(0, 1000000)
        # for debugging
        self.revert = revert

    # def _write_broken_patch(self):
    #     with open(
    #         f"log/failed_patches/patch_{self.failed_id}.diff", "w+", encoding="utf-8"
    #     ) as f:
    #         f.write(self.patch)

    def __enter__(self):
        if not self.patch:
            return

        try:
            if isinstance(self.patch, PatchFile):
                with open(self.patch.path, "w", encoding="utf-8") as f:
                    f.write(self.patch.patch)
            elif isinstance(self.patch, str):
                self.repo.apply_patch(self.patch)

        except GitCommandError as e:
            # self._write_broken_patch()
            raise PatchApplyExcepion(e)

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.patch:
            return

        try:
            if isinstance(self.patch, PatchFile) and self.revert:
                self.repo.reset_to_commit(self.head_commit)
            elif isinstance(self.patch, str) and self.revert:
                self.repo.reverse_patch(self.patch)

        except GitCommandError as e:
            log.info(f"Error reversing patch")
            raise PatchApplyExcepion(e)


class ResetLocalCommitContext:
    """
    Resets the repository to a specific commit
    """

    def __init__(self, repo: GitRepo, fd_reset: bool = False, revert: bool = True):
        self.repo = repo
        self.revert = revert
        self.fd_reset = fd_reset

    def __enter__(self):
        """
        Saves the current commit hash when entering the context.
        """
        self.original_commit = self.repo.head.commit.hexsha
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Restores the repository to the original commit when exiting the context.
        """
        if exc_type is not None:
            print(f"An exception occurred: {exc_type.__name__}: {exc_value}")
            # Optionally, log the traceback here

        if self.fd_reset:
            self._add_files()

        if self.original_commit and self.revert:
            self.repo.reset_to_commit(self.original_commit)

        # reset the patched files in GitRepo
        self.repo.patched_files = {}

    def _add_files(self):
        """
        Adds all files in repo so that they are tracked by git and can be resetted
        when the context is exited
        """
        self.repo.repo.git.add(".")
        print(self.repo.repo.git.status())

    def reset_to_commit(self, commit_sha: str, parent=None):
        """
        Resets the index of the repository to a specific commit.
        """
        return self.repo.reset_to_commit(commit_sha, parent)


class ResetRemoteCommitContext:
    """
    Pushes a single commit to remote repo and then reverts it in the remote
    by pushing HEAD~1. Used for unit tests
    """

    def __init__(self, repo: GitRepo):
        self.repo = repo
        self._called = False

    def __enter__(self):
        """
        Saves the current commit hash when entering the context.
        """
        self.original_commit = self.repo.head.commit.hexsha
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Restores remote repo to HEAD~1
        """
        if not self._called:
            # reset local repo state if we dont end up pushing
            self.repo.reset_to_commit(self.original_commit)
            print("Resetting to og state")
            return

        self.repo.reset_to_commit_head(head=1)
        self.repo.push()

    def add_commit_push(self, files: List[str], msg: str):
        """
        Adds a commit to the repo and pushes it to the remote
        """
        try:
            self.repo.add_n_commit(files, msg)
            self.repo.push()
        except Exception as e:
            print("Error adding commit")
            self.repo.reset_to_commit(self.original_commit)
            raise e
        finally:
            self._called = True



from setuptools import Extension, find_packages, setup

# TODO: figure out how to run npm run build
setup(
    name="cowboy-client",
    version="0.0.18",
    packages=find_packages(),
    description="Cowboy Client Interface",
    # long_description=long_description,
    # long_description_content_type="text/markdown",
    author="John Peng",
    author_email="kongyijipeng@gmail.com",
    install_requires=[
        "click",
        "pyyaml",
        "pydantic==2.7.4",
        "gitpython==3.0.6",
        "requests",
        "platformdirs",
        "pytz",
    ],
    entry_points={
        "console_scripts": [
            "cowboy = cowboy.cli:entrypoint",
        ],
    },
    python_requires=">=3.8",
    include_package_data=True,
    exclude_package_data={
        "": ["static/*"],
    },
)



from cowboy.utils import start_daemon, is_port_available

from http.server import SimpleHTTPRequestHandler
from webbrowser import open
import socketserver
import os
import sys

DIRECTORY = "build"
HOST = "localhost"
PORT = 8085
MAX_RETRIES = 10


class CustomHandler(SimpleHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=DIRECTORY, **kwargs)

    # need this to route all requests to index.html or else the web server
    # will default try look for files in the served directory
    def do_GET(self):
        if self.path != "/" and self.path != "" and self.path != "/index.html":
            if os.path.exists(self.translate_path(self.path)):
                print("Routing to file: ", self.path)
                return super().do_GET()

        self.path = "/"
        return super().do_GET()


def serve_ui(session_id):
    def run_server(port):
        print("Running server on port: ", port)
        with socketserver.TCPServer(("", port), CustomHandler) as httpd:
            httpd.serve_forever()

    retries = 0
    port = PORT

    while retries < MAX_RETRIES:
        if is_port_available(port):
            # technically doing this out of order but its fine ..
            open(f"http://{HOST}:{PORT}/test-results/{session_id}")
            start_daemon(run_server, (port,))
            return

        retries += 1
        port += 1



from cowboy_lib.repo.repository import PatchFileContext, GitRepo
from cowboy_lib.coverage import CoverageResult
from cowboy_lib.api.runner.shared import RunTestTaskArgs, FunctionArg
from cowboy.repo.models import RepoConfig
from cowboy.exceptions import CowboyClientError

from .base import RunnerError, TestSuiteError

import os
import subprocess
from typing import List, Tuple, NewType, Dict
import json
import hashlib
from pathlib import Path

from logging import getLogger

log = getLogger(__name__)


COVERAGE_FILE = "coverage.json"
TestError = NewType("TestError", str)


class DiffFileCreation(Exception):
    pass


def hash_str(s: str) -> str:
    return hashlib.sha256(s.encode()).hexdigest()


def hash_file(filepath):
    """Compute SHA-256 hash of the specified file"""
    with open(filepath, "r", encoding="utf-8") as f:
        buf = f.read()

    return hash_str(buf)


def hash_coverage_inputs(directory: Path, cmd_str: str) -> str:
    """Compute SHA-256 for the curr dir and cmd_str"""
    hashes = []
    for f in directory.iterdir():
        if f.is_file() and f.name.endswith(".py"):
            file_hash = hash_file(f)
            hashes.append((str(f), file_hash))

    # Sort based on file path and then combine the file hashes
    hashes.sort()
    combined_hash = hashlib.sha256()
    for _, file_hash in hashes:
        combined_hash.update(file_hash.encode())

    combined_hash.update(cmd_str.encode())
    return combined_hash.hexdigest()


from contextlib import contextmanager
import queue


class LockedRepos:
    """
    A list of available repos for concurrent run_test invocations, managed as a FIFO queue
    """

    def __init__(self, cloned_folders: List[Path]):
        self.capacity = len(cloned_folders)
        self.queue = queue.Queue()

        for path in cloned_folders:
            if not path.exists():
                raise RunnerError(
                    f"Cloned folder: {str(path)} does not exist, something went wrong. Try deleting and re-creating the repo"
                )

            self.queue.put(GitRepo(path))

    @contextmanager
    def acquire_one(self) -> GitRepo:
        git_repo: GitRepo = self.queue.get()  # This will block if the queue is empty

        log.info(f"Acquired repo: {git_repo.repo_folder}")
        try:
            yield git_repo
        finally:
            self.release(git_repo)

            log.info(f"Released repo: {git_repo.repo_folder}")

    def release(self, git_repo: GitRepo):
        self.queue.put(git_repo)  # Return the repo back to the queue

    def __len__(self):
        return self.queue.qsize()


def get_exclude_path(
    func: FunctionArg,
    rel_fp: Path,
):
    """
    Converts a Function path
    """
    excl_name = (
        (func.name.split(".")[0] + "::" + func.name.split(".")[1])
        if func.is_meth
        else func.name
    )

    # need to do this on windows
    return str(rel_fp).replace("\\", "/") + "::" + excl_name


class PytestDiffRunner:
    """
    Executes the test suite
    """

    # _runner_instances = {}

    # def __new__(cls, repo_conf: RepoConfig, test_suite: str = ""):
    #     """
    #     Retrieve or create an instance of PytestDiffRunner for the given repo_name.
    #     """
    #     if repo_conf.repo_name not in cls._runner_instances:
    #         cls._runner_instances[repo_conf.repo_name] = object.__new__(cls)
    #     return cls._runner_instances[repo_conf.repo_name]

    def __init__(
        self,
        # assume to be a test file for now
        repo_conf: RepoConfig,
        test_suite: str = "",
    ):
        self.src_folder = Path(repo_conf.source_folder)
        self.test_folder = Path(repo_conf.python_conf.test_folder)
        self.interpreter = Path(repo_conf.python_conf.interp)
        self.python_path = Path(repo_conf.python_conf.pythonpath)

        if not self.interpreter.exists():
            raise RunnerError(f"Runtime path {self.interpreter} does not exist")

        missing = self.check_missing_deps(self.interpreter)
        if missing:
            raise RunnerError(f"{missing}")

        self.cov_folders = [Path(p) for p in repo_conf.python_conf.cov_folders]
        cloned_folders = [Path(p) for p in repo_conf.cloned_folders]
        self.test_repos = LockedRepos(cloned_folders)

        if len(self.test_repos) == 0:
            raise CowboyClientError("No cloned repos created, perhaps run init again?")

        self.test_suite = test_suite

    def check_missing_deps(self, interp) -> List[bool]:
        """
        Check if test depdencies are installed in the interpreter
        """
        not_installed = []
        deps = ["pytest-cov", "pytest"]
        try:
            for dep in deps:
                result = subprocess.run(
                    [interp, "-m", "pip", "show", dep],
                    capture_output=True,
                    text=True,
                )
                if result.returncode != 0:
                    not_installed.append(dep)

        except (subprocess.CalledProcessError, FileNotFoundError):
            not_installed.append(dep)

        return (
            f"The following deps are not installed: {not_installed}"
            if not_installed
            else ""
        )

    def verify_clone_dirs(self, cloned_dirs: List[Path]):
        """
        Verifies that the hash of all *.py files are the same for each cloned dir
        """
        import hashlib

        hashes = []
        for clone in cloned_dirs:
            f_buf = ""
            for py_file in clone.glob("test*.py"):
                with open(py_file, "r") as f:
                    f_buf += f.read()

            f_buf_hash = hashlib.md5(f_buf.encode()).hexdigest()
            hashes.append(f_buf_hash)

        if any(h != hashes[0] for h in hashes):
            raise CowboyClientError("Cloned directories are not the same")

    def set_test_repos(self, repo_paths: List[Path]):
        self.test_repos = LockedRepos(
            list(zip(repo_paths, [GitRepo(p) for p in repo_paths]))
        )

        self.verify_clone_dirs(repo_paths)

    def _get_exclude_tests_arg_str(
        self, excluded_tests: List[Tuple[FunctionArg, Path]], cloned_path: Path
    ):
        """
        Convert the excluded tests into Pytest deselect args
        """
        if not excluded_tests:
            return ""

        # PATCH: just concatenate the path with base folder
        tranf_paths = []
        for test, test_fp in excluded_tests:
            tranf_paths.append(get_exclude_path(test, test_fp))

        return "--deselect=" + " --deselect=".join(tranf_paths)

    def _get_include_tests_arg_str(self, include_tests: []):
        if not include_tests:
            return ""

        arg_str = ""
        AND = " and"
        for test in include_tests:
            arg_str += f"{test}{AND}"

        arg_str = arg_str[: -len(AND)]
        # return "-k " + '"' + arg_str + '"'
        return "-k " + arg_str

    def _construct_cmd(
        self, repo_path, selected_tests: str = "", deselected_tests: str = ""
    ):
        """
        Constructs the cmdstr for running pytest
        """

        cmd = [
            "cd",
            str(repo_path),
            "&&",
            str(self.interpreter),
            "-m",
            "pytest",
            str(self.test_folder),
            "--tb",
            "short",
            selected_tests,
            deselected_tests,
            # "-v",
            "--color",
            "no",
            f"--cov={'--cov='.join([str(folder) + ' ' for folder in self.cov_folders])}",
            "--cov-report",
            "json",
            # "--cov-report",
            # "term",
            "--continue-on-collection-errors",
            "--disable-warnings",
        ]

        return " ".join(cmd)

    def run_testsuite(self, args: RunTestTaskArgs) -> Tuple[CoverageResult, str, str]:
        with self.test_repos.acquire_one() as repo_inst:
            git_repo: GitRepo = repo_inst

            patch_file = args.patch_file
            if patch_file:
                patch_file.path = git_repo.repo_folder / patch_file.path
                log.info(f"Using patch file: {patch_file.path}")

            exclude_tests = args.exclude_tests
            include_tests = args.include_tests

            env = os.environ.copy()
            if self.python_path:
                env["PYTHONPATH"] = self.python_path

            exclude_tests = self._get_exclude_tests_arg_str(
                exclude_tests, git_repo.repo_folder
            )
            include_tests = self._get_include_tests_arg_str(include_tests)
            cmd_str = self._construct_cmd(
                git_repo.repo_folder, include_tests, exclude_tests
            )

            log.info(f"Running with command: {cmd_str}")

            with PatchFileContext(git_repo, patch_file):
                proc = subprocess.Popen(
                    cmd_str,
                    # env=env,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,
                    text=True,
                )
                stdout, stderr = proc.communicate()

                # raise exception only no coverage
                try:
                    with open(git_repo.repo_folder / COVERAGE_FILE, "r") as f:
                        coverage_json = json.loads(f.read())
                        cov = CoverageResult(stdout, stderr, coverage_json)
                        if not cov.coverage:
                            print("No coverage!")
                            raise Exception()
                except Exception:
                    raise TestSuiteError(stderr, str(git_repo.repo_folder), cmd_str)

        return (
            cov,
            stdout,
            stderr,
        )

    # NOTE: keep this around, there is still that race condition bug we need
    # to find
    # def fake_test_suite(self, args: RunTestTaskArgs) -> Tuple[CoverageResult, str, str]:
    #     import time
    #     import random

    #     with self.test_repos.acquire_one() as repo_inst:
    #         print(f"Acquired repo: {repo_inst}")
    #         cloned_path, git_repo = repo_inst

    #         patch_file = args.patch_file
    #         if patch_file:
    #             patch_file.path = cloned_path / patch_file.path
    #             log.info(f"Using patch file: {patch_file.path}")

    #         exclude_tests = args.exclude_tests
    #         include_tests = args.include_tests

    #         env = os.environ.copy()
    #         if self.python_path:
    #             env["PYTHONPATH"] = self.python_path

    #         exclude_tests = self._get_exclude_tests_arg_str(exclude_tests, cloned_path)
    #         include_tests = self._get_include_tests_arg_str(include_tests)
    #         cmd_str = self._construct_cmd(cloned_path, include_tests, exclude_tests)
    #         print(f"Running with command: {cmd_str}")

    #         with PatchFileContext(git_repo, patch_file):
    #             time.sleep(random.randint(1, 10))
    #             # proc = subprocess.Popen(
    #             #     cmd_str,
    #             #     # env=env,
    #             #     stdout=subprocess.PIPE,
    #             #     stderr=subprocess.PIPE,
    #             #     shell=True,
    #             #     text=True,
    #             # )
    #             # stdout, stderr = proc.communicate()
    #             # if stderr:
    #             #     raise TestSuiteError(stderr)

    #             # # read pycoverage result
    #             # with open(cloned_path / COVERAGE_FILE, "r") as f:
    #             #     coverage_json = json.loads(f.read())
    #             #     cov = CoverageResult(stdout, stderr, coverage_json)

    #     return (
    #         None,
    #         "",
    #         "",
    #     )



import random
import string
import threading
from queue import Queue


def gen_random_name():
    """
    Generates a random name using ASCII, 8 characters in length
    """

    return "".join(random.choices(string.ascii_lowercase, k=8))


def start_daemon(task, args):
    """
    Starts a daemon thread that continuously joins/checks is_alive to
    allow for sigints to pass thru (which would otherwise get consumed
    by some blocking python functions)
    """
    result_queue = Queue()

    def target():
        try:
            result = task(*args)
            result_queue.put((None, result))
        except Exception as e:
            result_queue.put((e, None))

    t = threading.Thread(target=target, daemon=True)
    t.start()

    try:
        while t.is_alive():
            t.join(timeout=0.1)  # Allow signal handling
    except KeyboardInterrupt:
        raise KeyboardInterrupt

    exception, result = result_queue.get()
    if exception:
        raise exception

    return result


def is_port_available(port):
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("", port))
            return True
        except OSError:
            return False



from dataclasses import dataclass

from typing import List, Optional, TYPE_CHECKING, Tuple
from pathlib import Path

from cowboy_lib.ast.code import ASTNode


@dataclass
class TargetCode:
    """
    A chunk of code that is covered by the lines in a TestModule
    """

    range: Tuple[int, int]
    lines: List[str]
    filepath: Path
    func_scope: Optional[ASTNode]
    class_scope: Optional[ASTNode]

    def base_path(self) -> Path:
        """
        Returns the base path relative to the repo directory
        """
        return Path(*self.filepath.parts[2:])

    def __post_init__(self):
        if not isinstance(self.filepath, Path):
            self.filepath = Path(self.filepath)

    def __eq__(self, other: "TargetCode"):
        return self.filepath == other.filepath and self.range == other.range

    def __hash__(self):
        return hash((self.filepath, self.range))

    def to_lines(self):
        repr = ""
        for i, line in zip(range(self.range[0], self.range[1] + 1), self.lines):
            repr += f"{i}. {line}\n"

        return repr



from cowboy.config import LOG_DIR

import logging
import os
from datetime import datetime
import pytz


def converter(timestamp):
    dt = datetime.fromtimestamp(timestamp, tz=pytz.utc)
    return dt.astimezone(pytz.timezone("US/Eastern")).timetuple()


def get_file_handler(log_dir=LOG_DIR):
    """
    Returns a file handler for logging.
    """
    os.makedirs(log_dir, exist_ok=True)
    timestamp = datetime.now().strftime("%Y-%m-%d")
    file_name = f"runner_{timestamp}.log"
    file_handler = logging.FileHandler(os.path.join(log_dir, file_name))

    return file_handler


formatter = logging.Formatter(
    fmt="%(asctime)s - %(name)s:%(levelname)s: %(filename)s:%(lineno)s - %(message)s",
    datefmt="%H:%M:%S",
)
formatter.converter = converter

# converter can only be set on the formatter
file_handler = get_file_handler()
file_handler.setFormatter(formatter)


logging.basicConfig(
    level=logging.INFO,
    handlers=[file_handler],
)



from cowboy.http import APIClient
from cowboy.db.core import Database
from typing import List

from enum import Enum

db = Database()
api = APIClient(db)


class AugmentTestMode(str, Enum):
    AUTO = "auto"
    FILE = "files"
    TM = "module"
    ALL = "all"


def api_augment(repo_name: str, mode: str = "auto", tms: str = ""):
    """
    Augments existing test modules with new test cases
    """
    if mode not in [e.value for e in AugmentTestMode]:
        raise ValueError(
            f"Invalid mode {mode}, following are allowed: {', '.join(AugmentTestMode.__members__)}"
        )

    response = api.long_post(
        f"/test-gen/augment",
        {
            "repo_name": repo_name,
            "mode": mode,
            "tms": tms,
            "files": [],
        },
    )

    return response["session_id"]



import subprocess
import time
import signal
import sys

# List to hold the subprocesses
processes = []


def signal_handler(sig, frame):
    print("Main thread received interrupt signal. Terminating subprocesses...")
    for proc in processes:
        proc.terminate()  # Send SIGTERM
        try:
            proc.wait(timeout=5)  # Wait for the process to terminate
        except subprocess.TimeoutExpired:
            proc.kill()  # Force kill if not terminated
    sys.exit(0)


if __name__ == "__main__":
    import sys

    num_clients = int(sys.argv[1])
    if not num_clients:
        num_clients = 10

    # Register the signal handler for KeyboardInterrupt (Ctrl+C)
    signal.signal(signal.SIGINT, signal_handler)

    try:
        # Start 10 long-running subprocesses
        for i in range(num_clients):
            proc = subprocess.Popen(
                ["python", "-m", "cowboy.task_client.runtest_client", "hearbeat.txt", "1"],
                stdout=None,
                stderr=None,
            )
            processes.append(proc)
            print(f"Started subprocess {i} with PID {proc.pid}")

        # Keep the main thread running to maintain the relationship
        while True:
            time.sleep(1)

    except KeyboardInterrupt:
        signal_handler(None, None)
